"""
This module implements routines from reading and writing LEGEND Data Objects in
HDF5 files.
"""
from __future__ import annotations

import fnmatch
import glob
import logging
import os
import sys
from bisect import bisect_left
from collections import defaultdict
from typing import Any, Iterator, Union

import h5py
import numba as nb
import numpy as np
import pandas as pd

from . import compression as compress
from .array import Array
from .arrayofequalsizedarrays import ArrayOfEqualSizedArrays
from .compression import WaveformCodec
from .encoded import ArrayOfEncodedEqualSizedArrays, VectorOfEncodedVectors
from .fixedsizearray import FixedSizeArray
from .lgdo_utils import expand_path, parse_datatype
from .scalar import Scalar
from .struct import Struct
from .table import Table
from .vectorofvectors import VectorOfVectors
from .waveform_table import WaveformTable

LGDO = Union[Array, Scalar, Struct, VectorOfVectors]

log = logging.getLogger(__name__)

DEFAULT_HDF5_COMPRESSION = None


class LH5Store:
    """
    Class to represent a store of LEGEND HDF5 files. The two main methods
    implemented by the class are :meth:`read_object` and :meth:`write_object`.

    Examples
    --------
    >>> from pygama.lgdo import LH5Store
    >>> store = LH5Store()
    >>> obj, _ = store.read_object("/geds/waveform", "file.lh5")
    >>> type(obj)
    pygama.lgdo.waveform_table.WaveformTable
    """

    def __init__(self, base_path: str = "", keep_open: bool = False) -> None:
        """
        Parameters
        ----------
        base_path
            directory path to prepend to LH5 files.
        keep_open
            whether to keep files open by storing the :mod:`h5py` objects as
            class attributes.
        """
        self.base_path = "" if base_path == "" else expand_path(base_path)
        self.keep_open = keep_open
        self.files = {}

    def gimme_file(self, lh5_file: str | h5py.File, mode: str = "r") -> h5py.File:
        """Returns a :mod:`h5py` file object from the store or creates a new one.

        Parameters
        ----------
        lh5_file
            LH5 file name.
        mode
            mode in which to open file. See :class:`h5py.File` documentation.
        """
        if isinstance(lh5_file, h5py.File):
            return lh5_file
        if mode == "r":
            lh5_file = expand_path(lh5_file, base_path=self.base_path)
        if lh5_file in self.files.keys():
            return self.files[lh5_file]
        if self.base_path != "":
            full_path = os.path.join(self.base_path, lh5_file)
        else:
            full_path = lh5_file
        if mode != "r":
            directory = os.path.dirname(full_path)
            if directory != "" and not os.path.exists(directory):
                log.debug(f"making path {directory}")
                os.makedirs(directory)
        if mode == "r" and not os.path.exists(full_path):
            raise FileNotFoundError(f"file {full_path} not found")
        if mode != "r" and os.path.exists(full_path):
            log.debug(f"opening existing file {full_path} in mode '{mode}'")
        h5f = h5py.File(full_path, mode)
        if self.keep_open:
            self.files[lh5_file] = h5f
        return h5f

    def gimme_group(
        self,
        group: str | h5py.Group,
        base_group: h5py.Group,
        grp_attrs: dict[str, Any] = None,
        overwrite: bool = False,
    ) -> h5py.Group:
        """
        Returns an existing :class:`h5py` group from a base group or creates a
        new one. Can also set (or replace) group attributes.

        Parameters
        ----------
        group
            name of the HDF5 group.
        base_group
            HDF5 group to be used as a base.
        grp_attrs
            HDF5 group attributes.
        overwrite
            whether overwrite group attributes, ignored if `grp_attrs` is
            ``None``.
        """
        if not isinstance(group, h5py.Group):
            if group in base_group:
                group = base_group[group]
            else:
                group = base_group.create_group(group)
                if grp_attrs is not None:
                    group.attrs.update(grp_attrs)
                return group
        if (
            grp_attrs is not None
            and len(set(grp_attrs.items()) ^ set(group.attrs.items())) > 0
        ):
            if not overwrite:
                raise RuntimeError("grp_attrs != group.attrs but overwrite not set")
            else:
                log.debug(f"overwriting {group}.attrs...")
                for key in group.attrs.keys():
                    group.attrs.pop(key)
                group.attrs.update(grp_attrs)
        return group

    def get_buffer(
        self,
        name: str,
        lh5_file: str | h5py.File | list[str | h5py.File],
        size: int = None,
        field_mask: dict[str, bool] | list[str] | tuple[str] = None,
    ) -> LGDO:
        """Returns an LH5 object appropriate for use as a pre-allocated buffer
        in a read loop. Sets size to `size` if object has a size.
        """
        obj, n_rows = self.read_object(name, lh5_file, n_rows=0, field_mask=field_mask)
        if hasattr(obj, "resize") and size is not None:
            obj.resize(new_size=size)
        return obj

    def read_object(
        self,
        name: str,
        lh5_file: str | h5py.File | list[str | h5py.File],
        start_row: int = 0,
        n_rows: int = sys.maxsize,
        idx: np.ndarray | list | tuple | list[np.ndarray | list | tuple] = None,
        field_mask: dict[str, bool] | list[str] | tuple[str] = None,
        obj_buf: LGDO = None,
        obj_buf_start: int = 0,
        decompress: bool = True,
    ) -> tuple[LGDO, int]:
        """Read LH5 object data from a file.

        Parameters
        ----------
        name
            Name of the LH5 object to be read (including its group path).
        lh5_file
            The file(s) containing the object to be read out. If a list of
            files, array-like object data will be concatenated into the output
            object.
        start_row
            Starting entry for the object read (for array-like objects). For a
            list of files, only applies to the first file.
        n_rows
            The maximum number of rows to read (for array-like objects). The
            actual number of rows read will be returned as one of the return
            values (see below).
        idx
            For NumPy-style "fancying indexing" for the read. Used to read out
            rows that pass some selection criteria. Only selection along the first
            axis is supported, so tuple arguments must be one-tuples.  If `n_rows`
            is not false, `idx` will be truncated to `n_rows` before reading. To use
            with a list of files, can pass in a list of `idx`'s (one for each
            file) or use a long contiguous list (e.g. built from a previous
            identical read). If used in conjunction with `start_row` and `n_rows`,
            will be sliced to obey those constraints, where `n_rows` is
            interpreted as the (max) number of *selected* values (in `idx`) to be
            read out.
        field_mask
            For tables and structs, determines which fields get written out.
            Only applies to immediate fields of the requested objects. If a dict
            is used, a default dict will be made with the default set to the
            opposite of the first element in the dict. This way if one specifies
            a few fields at ``False``, all but those fields will be read out,
            while if one specifies just a few fields as ``True``, only those
            fields will be read out. If a list is provided, the listed fields
            will be set to ``True``, while the rest will default to ``False``.
        obj_buf
            Read directly into memory provided in `obj_buf`. Note: the buffer
            will be expanded to accommodate the data requested. To maintain the
            buffer length, send in ``n_rows = len(obj_buf)``.
        obj_buf_start
            Start location in ``obj_buf`` for read. For concatenating data to
            array-like objects.
        decompress
            Decompress data encoded with pygama's compression routines right
            after reading. The option has no effect on data encoded with HDF5
            built-in filters, which is always decompressed upstream by HDF5.

        Returns
        -------
        (object, n_rows_read)
            `object` is the read-out object `n_rows_read` is the number of rows
            successfully read out. Essential for arrays when the amount of data
            is smaller than the object buffer.  For scalars and structs
            `n_rows_read` will be``1``. For tables it is redundant with
            ``table.loc``.
        """
        # Handle list-of-files recursively
        if not isinstance(lh5_file, (str, h5py.File)):
            lh5_file = list(lh5_file)
            n_rows_read = 0
            for i, h5f in enumerate(lh5_file):
                if isinstance(idx, list) and len(idx) > 0 and not np.isscalar(idx[0]):
                    # a list of lists: must be one per file
                    idx_i = idx[i]
                elif idx is not None:
                    # make idx a proper tuple if it's not one already
                    if not (isinstance(idx, tuple) and len(idx) == 1):
                        idx = (idx,)
                    # idx is a long continuous array
                    n_rows_i = self.read_n_rows(name, h5f)
                    # find the length of the subset of idx that contains indices
                    # that are less than n_rows_i
                    n_rows_to_read_i = bisect_left(idx[0], n_rows_i)
                    # now split idx into idx_i and the remainder
                    idx_i = (idx[0][:n_rows_to_read_i],)
                    idx = (idx[0][n_rows_to_read_i:] - n_rows_i,)
                else:
                    idx_i = None
                n_rows_i = n_rows - n_rows_read
                obj_buf, n_rows_read_i = self.read_object(
                    name,
                    lh5_file[i],
                    start_row=start_row,
                    n_rows=n_rows_i,
                    idx=idx_i,
                    field_mask=field_mask,
                    obj_buf=obj_buf,
                    obj_buf_start=obj_buf_start,
                    decompress=decompress,
                )
                n_rows_read += n_rows_read_i
                if n_rows_read >= n_rows or obj_buf is None:
                    return obj_buf, n_rows_read
                start_row = 0
                obj_buf_start += n_rows_read_i
            return obj_buf, n_rows_read

        # get the file from the store
        h5f = self.gimme_file(lh5_file, "r")
        if not h5f or name not in h5f:
            raise KeyError(f"'{name}' not in {h5f.filename}")

        log.debug(
            f"reading {h5f.filename}:{name}[{start_row}:{n_rows}], decompress = {decompress}, "
            + (f" with field mask {field_mask}" if field_mask else "")
        )

        # make idx a proper tuple if it's not one already
        if not (isinstance(idx, tuple) and len(idx) == 1):
            if idx is not None:
                idx = (idx,)

        # get the object's datatype
        if "datatype" not in h5f[name].attrs:
            raise RuntimeError(
                f"'{name}' in file {lh5_file} is missing the datatype attribute"
            )

        datatype = h5f[name].attrs["datatype"]
        datatype, shape, elements = parse_datatype(datatype)

        # check field_mask and make it a default dict
        if datatype == "struct" or datatype == "table":
            if field_mask is None:
                field_mask = defaultdict(lambda: True)
            elif isinstance(field_mask, dict):
                default = True
                if len(field_mask) > 0:
                    default = not field_mask[list(field_mask.keys())[0]]
                field_mask = defaultdict(lambda: default, field_mask)
            elif isinstance(field_mask, (list, tuple)):
                field_mask = defaultdict(
                    lambda: False, {field: True for field in field_mask}
                )
            elif not isinstance(field_mask, defaultdict):
                raise RuntimeError("bad field_mask of type", type(field_mask).__name__)
        elif field_mask is not None:
            raise RuntimeError(f"datatype {datatype} does not accept a field_mask")

        # Scalar
        # scalars are dim-0 datasets
        if datatype == "scalar":
            value = h5f[name][()]
            if elements == "bool":
                value = np.bool_(value)
            if obj_buf is not None:
                obj_buf.value = value
                obj_buf.attrs.update(h5f[name].attrs)
                return obj_buf, 1
            else:
                return Scalar(value=value, attrs=h5f[name].attrs), 1

        # Struct
        # recursively build a struct, return as a dictionary
        if datatype == "struct":
            # ignore obj_buf.
            # TODO: could append new fields or overwrite/concat to existing
            # fields. If implemented, get_buffer() above should probably also
            # (optionally?) prep buffers for each field
            if obj_buf is not None:
                raise NotImplementedError("obj_buf not implemented for LGOD Structs")

            # loop over fields and read
            obj_dict = {}
            for field in elements:
                if not field_mask[field]:
                    continue
                # TODO: it's strange to pass start_row, n_rows, idx to struct
                # fields. If they all had shared indexing, they should be in a
                # table... Maybe should emit a warning? Or allow them to be
                # dicts keyed by field name?
                if "int_keys" in h5f[name].attrs:
                    if dict(h5f[name].attrs)["int_keys"]:
                        f = int(field)
                else:
                    f = str(field)
                obj_dict[f], _ = self.read_object(
                    name + "/" + field,
                    h5f,
                    start_row=start_row,
                    n_rows=n_rows,
                    idx=idx,
                    decompress=decompress,
                )
            # modify datatype in attrs if a field_mask was used
            attrs = dict(h5f[name].attrs)
            if field_mask is not None:
                selected_fields = []
                for field in elements:
                    if field_mask[field]:
                        selected_fields.append(field)
                attrs["datatype"] = "struct" + "{" + ",".join(selected_fields) + "}"
            return Struct(obj_dict=obj_dict, attrs=attrs), 1

        # Below here is all array-like types. So trim idx if needed
        if idx is not None:
            # chop off indices < start_row
            i_first_valid = bisect_left(idx[0], start_row)
            idxa = idx[0][i_first_valid:]
            # don't readout more than n_rows indices
            idx = (idxa[:n_rows],)  # works even if n_rows > len(idxa)

        # Table or WaveformTable
        if datatype == "table":
            col_dict = {}

            # read out each of the fields
            rows_read = []
            for field in elements:
                if not field_mask[field]:
                    continue

                fld_buf = None
                if obj_buf is not None:
                    if not isinstance(obj_buf, Table) or field not in obj_buf:
                        raise ValueError(
                            f"obj_buf for LGDO Table '{name}' not formatted correctly"
                        )

                    else:
                        fld_buf = obj_buf[field]

                col_dict[field], n_rows_read = self.read_object(
                    name + "/" + field,
                    h5f,
                    start_row=start_row,
                    n_rows=n_rows,
                    idx=idx,
                    obj_buf=fld_buf,
                    obj_buf_start=obj_buf_start,
                    decompress=decompress,
                )
                if obj_buf is not None and obj_buf_start + n_rows_read > len(obj_buf):
                    obj_buf.resize(obj_buf_start + n_rows_read)

                rows_read.append(n_rows_read)

            # warn if all columns don't read in the same number of rows
            if len(rows_read) > 0:
                n_rows_read = rows_read[0]
            else:
                n_rows_read = 0
                log.warning(f"Table '{name}' has no subgroups accepted by field mask")

            for n in rows_read[1:]:
                if n != n_rows_read:
                    log.warning(
                        f"Table '{name}' got strange n_rows_read = {n}, {n_rows_read} was expected ({rows_read})"
                    )

            # modify datatype in attrs if a field_mask was used
            attrs = dict(h5f[name].attrs)
            if field_mask is not None:
                selected_fields = []
                for field in elements:
                    if field_mask[field]:
                        selected_fields.append(field)
                attrs["datatype"] = "table" + "{" + ",".join(selected_fields) + "}"

            # fields have been read out, now return a table
            if obj_buf is None:
                # if col_dict contains just 3 objects called t0, dt, and values,
                # return a WaveformTable
                if (
                    len(col_dict) == 3
                    and "t0" in col_dict
                    and "dt" in col_dict
                    and "values" in col_dict
                ):
                    table = WaveformTable(
                        t0=col_dict["t0"], dt=col_dict["dt"], values=col_dict["values"]
                    )
                else:
                    table = Table(col_dict=col_dict, attrs=attrs)

                # set (write) loc to end of tree
                table.loc = n_rows_read
                return table, n_rows_read
            else:
                # We have read all fields into the object buffer. Run
                # checks: All columns should be the same size. So update
                # table's size as necessary, warn if any mismatches are found
                obj_buf.resize(do_warn=True)
                # set (write) loc to end of tree
                obj_buf.loc = obj_buf_start + n_rows_read
                # check attributes
                if set(obj_buf.attrs.keys()) != set(attrs.keys()):
                    raise RuntimeError(
                        f"attrs mismatch. obj_buf.attrs: "
                        f"{obj_buf.attrs}, h5f[{name}].attrs: {attrs}"
                    )
                return obj_buf, n_rows_read

        # ArrayOfEncodedEqualSizedArrays and VectorOfEncodedVectors
        for cond, enc_lgdo in [
            (
                datatype == "array_of_encoded_equalsized_arrays",
                ArrayOfEncodedEqualSizedArrays,
            ),
            (elements.startswith("encoded_array"), VectorOfEncodedVectors),
        ]:
            if cond:
                if (
                    not decompress
                    and obj_buf is not None
                    and not isinstance(obj_buf, enc_lgdo)
                ):
                    raise ValueError(f"obj_buf for '{name}' not a {enc_lgdo}")

                # read out decoded_size, either a Scalar or an Array
                decoded_size_buf = encoded_data_buf = None
                if obj_buf is not None and not decompress:
                    decoded_size_buf = obj_buf.decoded_size
                    encoded_data_buf = obj_buf.encoded_data

                decoded_size, _ = self.read_object(
                    f"{name}/decoded_size",
                    h5f,
                    start_row=start_row,
                    n_rows=n_rows,
                    idx=idx,
                    obj_buf=None if decompress else decoded_size_buf,
                    obj_buf_start=0 if decompress else obj_buf_start,
                )

                # read out encoded_data, a VectorOfVectors
                encoded_data, n_rows_read = self.read_object(
                    f"{name}/encoded_data",
                    h5f,
                    start_row=start_row,
                    n_rows=n_rows,
                    idx=idx,
                    obj_buf=None if decompress else encoded_data_buf,
                    obj_buf_start=0 if decompress else obj_buf_start,
                )

                # return the still encoded data in the buffer object, if there
                if obj_buf is not None and not decompress:
                    return obj_buf, n_rows_read

                # otherwise re-create the encoded LGDO
                rawdata = enc_lgdo(
                    encoded_data=encoded_data,
                    decoded_size=decoded_size,
                    attrs=h5f[name].attrs,
                )

                # already return if no decompression is requested
                if not decompress:
                    return rawdata, n_rows_read

                # if no buffer, decode and return
                elif obj_buf is None and decompress:
                    return compress.decode(rawdata), n_rows_read

                # use the (decoded object type) buffer otherwise
                if enc_lgdo == VectorOfEncodedVectors and not isinstance(
                    obj_buf, VectorOfVectors
                ):
                    raise ValueError(
                        f"obj_buf for decoded '{name}' not a VectorOfVectors"
                    )
                elif enc_lgdo == ArrayOfEncodedEqualSizedArrays and not isinstance(
                    obj_buf, ArrayOfEqualSizedArrays
                ):
                    raise ValueError(
                        f"obj_buf for decoded '{name}' not an ArrayOfEqualSizedArrays"
                    )

                # FIXME: not a good idea. an in place decoding version
                # of decode would be needed to avoid extra memory
                # allocations
                # FIXME: obj_buf_start??? Write a unit test
                for i, wf in enumerate(compress.decode(rawdata)):
                    obj_buf[i] = wf

                return obj_buf, n_rows_read

        # VectorOfVectors
        # read out vector of vectors of different size
        if elements.startswith("array"):
            if obj_buf is not None and not isinstance(obj_buf, VectorOfVectors):
                raise ValueError(f"obj_buf for '{name}' not a LGDO VectorOfVectors")

            # read out cumulative_length
            cumulen_buf = None if obj_buf is None else obj_buf.cumulative_length
            cumulative_length, n_rows_read = self.read_object(
                f"{name}/cumulative_length",
                h5f,
                start_row=start_row,
                n_rows=n_rows,
                idx=idx,
                obj_buf=cumulen_buf,
                obj_buf_start=obj_buf_start,
            )
            # get a view of just what was read out for cleaner code below
            this_cumulen_nda = cumulative_length.nda[
                obj_buf_start : obj_buf_start + n_rows_read
            ]

            if idx is not None and n_rows_read > 0:
                # get the starting indices for each array in flattended data:
                # the starting index for array[i] is cumulative_length[i-1]
                idx2 = (np.asarray(idx[0]).copy() - 1,)
                # re-read cumulative_length with these indices
                # note this will allocate memory for fd_starts!
                fd_start = None
                if idx2[0][0] == -1:
                    idx2 = (idx2[0][1:],)
                    fd_start = 0  # this variable avoids an ndarray append
                fd_starts, fds_n_rows_read = self.read_object(
                    f"{name}/cumulative_length",
                    h5f,
                    start_row=start_row,
                    n_rows=n_rows,
                    idx=idx2,
                )
                fd_starts = fd_starts.nda  # we just need the nda
                if fd_start is None:
                    fd_start = fd_starts[0]

                # compute the length that flattened_data will have after the
                # fancy-indexed read
                fd_n_rows = np.sum(this_cumulen_nda[-len(fd_starts) :] - fd_starts)
                if fd_start == 0:
                    fd_n_rows += this_cumulen_nda[0]

                # now make fd_idx
                fd_idx = np.empty(fd_n_rows, dtype="uint32")
                fd_idx = _make_fd_idx(fd_starts, this_cumulen_nda, fd_idx)

                # Now clean up this_cumulen_nda, to be ready
                # to match the in-memory version of flattened_data. Note: these
                # operations on the view change the original array because they are
                # numpy arrays, not lists.
                this_cumulen_nda[-len(fd_starts) :] -= fd_starts
                np.cumsum(this_cumulen_nda, out=this_cumulen_nda)

            else:
                fd_idx = None

                # determine the start_row and n_rows for the flattened_data readout
                fd_start = 0
                if start_row > 0 and n_rows_read > 0:
                    # need to read out the cumulen sample -before- the first sample
                    # read above in order to get the starting row of the first
                    # vector to read out in flattened_data
                    fd_start = h5f[f"{name}/cumulative_length"][start_row - 1]

                    # check limits for values that will be used subsequently
                    if this_cumulen_nda[-1] < fd_start:
                        log.debug(
                            f"this_cumulen_nda[-1] = {this_cumulen_nda[-1]}, "
                            f"fd_start = {fd_start}, "
                            f"start_row = {start_row}, "
                            f"n_rows_read = {n_rows_read}"
                        )
                        raise RuntimeError(
                            f"cumulative_length non-increasing between entries "
                            f"{start_row} and {start_row+n_rows_read} ??"
                        )

                # determine the number of rows for the flattened_data readout
                fd_n_rows = this_cumulen_nda[-1] if n_rows_read > 0 else 0

                # Now done with this_cumulen_nda, so we can clean it up to be ready
                # to match the in-memory version of flattened_data. Note: these
                # operations on the view change the original array because they are
                # numpy arrays, not lists.
                #
                # First we need to subtract off the in-file offset for the start of
                # read for flattened_data
                this_cumulen_nda -= fd_start

            # If we started with a partially-filled buffer, add the
            # appropriate offset for the start of the in-memory flattened
            # data for this read.
            fd_buf_start = np.uint32(0)
            if obj_buf_start > 0:
                fd_buf_start = cumulative_length.nda[obj_buf_start - 1]
                this_cumulen_nda += fd_buf_start

            # Now prepare the object buffer if necessary
            fd_buf = None
            if obj_buf is not None:
                fd_buf = obj_buf.flattened_data
                # grow fd_buf if necessary to hold the data
                fdb_size = fd_buf_start + fd_n_rows
                if len(fd_buf) < fdb_size:
                    fd_buf.resize(fdb_size)

            # now read
            flattened_data, dummy_rows_read = self.read_object(
                f"{name}/flattened_data",
                h5f,
                start_row=fd_start,
                n_rows=fd_n_rows,
                idx=fd_idx,
                obj_buf=fd_buf,
                obj_buf_start=fd_buf_start,
            )
            if obj_buf is not None:
                return obj_buf, n_rows_read
            return (
                VectorOfVectors(
                    flattened_data=flattened_data,
                    cumulative_length=cumulative_length,
                    attrs=h5f[name].attrs,
                ),
                n_rows_read,
            )

        # Array
        # FixedSizeArray
        # ArrayOfEqualSizedArrays
        # read out all arrays by slicing
        if "array" in datatype:
            if obj_buf is not None:
                if not isinstance(obj_buf, Array):
                    raise ValueError(f"obj_buf for '{name}' not an LGDO Array")
                    obj_buf = None

            # compute the number of rows to read
            # we culled idx above for start_row and n_rows, now we have to apply
            # the constraint of the length of the dataset
            ds_n_rows = h5f[name].shape[0]
            if idx is not None:
                if len(idx[0]) > 0 and idx[0][-1] >= ds_n_rows:
                    log.warning(
                        "idx indexed past the end of the array in the file. Culling..."
                    )
                    n_rows_to_read = bisect_left(idx[0], ds_n_rows)
                    idx = (idx[0][:n_rows_to_read],)
                if len(idx[0]) == 0:
                    log.warning("idx empty after culling.")
                n_rows_to_read = len(idx[0])
            else:
                n_rows_to_read = ds_n_rows - start_row
            if n_rows_to_read > n_rows:
                n_rows_to_read = n_rows

            # prepare the selection for the read. Use idx if available
            if idx is not None:
                source_sel = idx
            else:
                source_sel = np.s_[start_row : start_row + n_rows_to_read]

            # Now read the array
            if obj_buf is not None and n_rows_to_read > 0:
                buf_size = obj_buf_start + n_rows_to_read
                if len(obj_buf) < buf_size:
                    obj_buf.resize(buf_size)
                dest_sel = np.s_[obj_buf_start:buf_size]
                h5f[name].read_direct(obj_buf.nda, source_sel, dest_sel)
                nda = obj_buf.nda
            else:
                if n_rows == 0:
                    tmp_shape = (0,) + h5f[name].shape[1:]
                    nda = np.empty(tmp_shape, h5f[name].dtype)
                else:
                    nda = h5f[name][source_sel]

            # special handling for bools
            # (c and Julia store as uint8 so cast to bool)
            if elements == "bool":
                nda = nda.astype(np.bool_)

            # Finally, set attributes and return objects
            attrs = h5f[name].attrs
            if obj_buf is None:
                if datatype == "array":
                    return Array(nda=nda, attrs=attrs), n_rows_to_read
                if datatype == "fixedsize_array":
                    return FixedSizeArray(nda=nda, attrs=attrs), n_rows_to_read
                if datatype == "array_of_equalsized_arrays":
                    return (
                        ArrayOfEqualSizedArrays(nda=nda, dims=shape, attrs=attrs),
                        n_rows_to_read,
                    )
            else:
                if set(obj_buf.attrs.keys()) != set(attrs.keys()):
                    raise RuntimeError(
                        f"attrs mismatch. "
                        f"obj_buf.attrs: {obj_buf.attrs}, "
                        f"h5f[{name}].attrs: {attrs}"
                    )
                return obj_buf, n_rows_to_read

        raise RuntimeError("don't know how to read datatype {datatype}")

    def write_object(
        self,
        obj: LGDO,
        name: str,
        lh5_file: str | h5py.File,
        group: str | h5py.Group = "/",
        start_row: int = 0,
        n_rows: int = None,
        wo_mode: str = "append",
        write_start: int = 0,
        hdf5_compression: str | h5py.filters.FilterRefBase = DEFAULT_HDF5_COMPRESSION,
    ) -> None:
        """Write an LGDO into an LH5 file.

        If the `obj` :class:`.LGDO` has a `compression` attribute, its value is
        interpreted as the algorithm to be used to compress `obj` before
        writing to disk. The type of `compression` can be:

        string, kwargs dictionary, hdf5plugin filter
          interpreted as the name of a built-in or custom `HDF5 compression
          filter <https://docs.h5py.org/en/stable/high/dataset.html#filter-pipeline>`_
          (``"gzip"``, ``"lzf"``, :mod:`hdf5plugin` filter object etc.) and
          passed directly to :meth:`h5py.Group.create_dataset`.

        :class:`.WaveformCodec` object
          If `obj` is a :class:`.WaveformTable`, compress its `values` using
          this algorithm. More documentation about the supported waveform
          compression algorithms at :mod:`.lgdo.compression`.

        Note
        ----
        The `compression` attribute takes precedence over the
        `hdf5_compression` argument and is not written to disk.

        Note
        ----
        HDF5 compression is skipped for the `encoded_data` dataset of
        :class:`.VectorOfEncodedVectors` and
        :class`.ArrayOfEncodedEqualSizedArrays`.

        Parameters
        ----------
        obj
            LH5 object. if object is array-like, writes `n_rows` starting from
            `start_row` in `obj`.
        name
            name of the object in the output HDF5 file.
        lh5_file
            HDF5 file name or :class:`h5py.File` object.
        group
            HDF5 group name or :class:`h5py.Group` object in which `obj` should
            be written.
        start_row
            first row in `obj` to be written.
        n_rows
            number of rows in `obj` to be written.
        wo_mode
            - ``write_safe`` or ``w``: only proceed with writing if the
              object does not already exist in the file.
            - ``append`` or ``a``: append along axis 0 (the first dimension)
              of array-like objects and array-like subfields of structs.
              :class:`~.lgdo.scalar.Scalar` objects get overwritten.
            - ``overwrite`` or ``o``: replace data in the file if present,
              starting from `write_start`. Note: overwriting with `write_start` =
              end of array is the same as ``append``.
            - ``overwrite_file`` or ``of``: delete file if present prior to
              writing to it. `write_start` should be 0 (its ignored).
            - ``append_column`` or ``ac``: append columns from an :class:`~.lgdo.table.Table`
              `obj` only if there is an existing :class:`~.lgdo.table.Table` in the `lh5_file` with
              the same `name` and :class:`~.lgdo.table.Table.size`. If the sizes don't match,
              or if there are matching fields, it errors out.
        write_start
            row in the output file (if already existing) to start overwriting
            from.
        hdf5_compression
            HDF5 compression filter to be applied before writing non-scalar
            datasets. **Ignored if compression is specified as an `obj`
            attribute.**
        """
        log.debug(
            f"writing {repr(obj)}[{start_row}:{n_rows}] as "
            f"{lh5_file}:{group}/{name}[{write_start}:], "
            f"mode = {wo_mode}, hdf5_compression = {hdf5_compression}"
        )

        if wo_mode == "write_safe":
            wo_mode = "w"
        if wo_mode == "append":
            wo_mode = "a"
        if wo_mode == "overwrite":
            wo_mode = "o"
        if wo_mode == "overwrite_file":
            wo_mode = "of"
            write_start = 0
        if wo_mode == "append_column":
            wo_mode = "ac"
        if wo_mode not in ["w", "a", "o", "of", "ac"]:
            raise ValueError(f"unknown wo_mode '{wo_mode}'")

        # "mode" is for the h5df.File and wo_mode is for this function
        # In hdf5, 'a' is really "modify" -- in addition to appending, you can
        # change any object in the file. So we use file:append for
        # write_object:overwrite.
        mode = "w" if wo_mode == "of" else "a"
        lh5_file = self.gimme_file(lh5_file, mode=mode)
        group = self.gimme_group(group, lh5_file)
        if wo_mode == "w" and name in group:
            raise RuntimeError(f"can't overwrite '{name}' in wo_mode 'write_safe'")

        # struct or table or waveform table
        if isinstance(obj, Struct):
            # In order to append a column, we need to update the `table{old_fields}` value in `group.attrs['datatype"]` to include the new fields.
            # One way to do this is to override `obj.attrs["datatype"]` to include old and new fields. Then we can write the fields to the table as normal.
            if wo_mode == "ac":
                old_group = self.gimme_group(name, group)
                datatype, shape, fields = parse_datatype(old_group.attrs["datatype"])
                if datatype not in ["table", "struct"]:
                    raise RuntimeError(
                        f"Trying to append columns to an object of type {datatype}"
                    )

                # If the mode is `append_column`, make sure we aren't appending a table that has a column of the same name as in the existing table
                # Also make sure that the field we are adding has the same size
                if len(list(set(fields).intersection(set(obj.keys())))) != 0:
                    raise ValueError(
                        f"Can't append {list(set(fields).intersection(set(obj.keys())))} column(s) to a table with the same field(s)"
                    )
                # It doesn't matter what key we access, as all fields in the old table have the same size
                if old_group[list(old_group.keys())[0]].size != obj.size:
                    raise ValueError(
                        f"Table sizes don't match. Trying to append column of size {obj.size} to a table of size {old_group[list(old_group.keys())[0]].size}."
                    )

                # Now we can append the obj.keys() to the old fields, and then update obj.attrs.
                fields.extend(list(obj.keys()))
                obj.attrs.pop("datatype")
                obj.attrs["datatype"] = "table" + "{" + ",".join(fields) + "}"

            group = self.gimme_group(
                name,
                group,
                grp_attrs=obj.attrs,
                overwrite=(wo_mode in ["o", "ac"]),
            )
            # If the mode is overwrite, then we need to peek into the file's table's existing fields
            # If we are writing a new table to the group that does not contain an old field, we should delete that old field from the file
            if wo_mode == "o":
                # Find the old keys in the group that are not present in the new table's keys, then delete them
                for key in list(set(group.keys()) - set(obj.keys())):
                    log.debug(f"{key} is not present in new table, deleting field")
                    del group[key]

            for field in obj.keys():
                # eventually compress waveform table values with pygama's
                # custom codecs before writing
                # if waveformtable.values.attrs["compression"] is a string,
                # interpret it as an HDF5 built-in filter
                obj_fld = None
                if (
                    isinstance(obj, WaveformTable)
                    and field == "values"
                    and not isinstance(obj.values, VectorOfEncodedVectors)
                    and not isinstance(obj.values, ArrayOfEncodedEqualSizedArrays)
                    and "compression" in obj.values.attrs
                    and isinstance(obj.values.attrs["compression"], WaveformCodec)
                ):
                    codec = obj.values.attrs["compression"]
                    obj_fld = compress.encode(obj.values, codec=codec)
                else:
                    obj_fld = obj[field]

                # Convert keys to string for dataset names
                f = str(field)
                self.write_object(
                    obj_fld,
                    f,
                    lh5_file,
                    group=group,
                    start_row=start_row,
                    n_rows=n_rows,
                    wo_mode=wo_mode,
                    write_start=write_start,
                    hdf5_compression=hdf5_compression,
                )
            return

        # scalars
        elif isinstance(obj, Scalar):
            if name in group:
                if wo_mode in ["o", "a"]:
                    log.debug(f"overwriting {name} in {group}")
                    del group[name]
                else:
                    raise RuntimeError(
                        f"tried to overwrite {name} in {group} for wo_mode {wo_mode}"
                    )
            ds = group.create_dataset(name, shape=(), data=obj.value)
            ds.attrs.update(obj.attrs)
            return

        # vector of encoded vectors
        elif isinstance(obj, (VectorOfEncodedVectors, ArrayOfEncodedEqualSizedArrays)):
            group = self.gimme_group(
                name, group, grp_attrs=obj.attrs, overwrite=(wo_mode == "o")
            )

            self.write_object(
                obj.encoded_data,
                "encoded_data",
                lh5_file,
                group=group,
                start_row=start_row,
                n_rows=n_rows,
                wo_mode=wo_mode,
                write_start=write_start,
                hdf5_compression=None,  # data is already compressed!
            )

            self.write_object(
                obj.decoded_size,
                "decoded_size",
                lh5_file,
                group=group,
                start_row=start_row,
                n_rows=n_rows,
                wo_mode=wo_mode,
                write_start=write_start,
                hdf5_compression=hdf5_compression,
            )

        # vector of vectors
        elif isinstance(obj, VectorOfVectors):
            group = self.gimme_group(
                name, group, grp_attrs=obj.attrs, overwrite=(wo_mode == "o")
            )
            if (
                n_rows is None
                or n_rows > obj.cumulative_length.nda.shape[0] - start_row
            ):
                n_rows = obj.cumulative_length.nda.shape[0] - start_row

            # if appending we need to add an appropriate offset to the
            # cumulative lengths as appropriate for the in-file object
            offset = 0  # declare here because we have to subtract it off at the end
            if (wo_mode == "a" or wo_mode == "o") and "cumulative_length" in group:
                len_cl = len(group["cumulative_length"])
                if wo_mode == "a":
                    write_start = len_cl
                if len_cl > 0:
                    offset = group["cumulative_length"][write_start - 1]

            # First write flattened_data array. Only write rows with data.
            fd_start = 0 if start_row == 0 else obj.cumulative_length.nda[start_row - 1]
            fd_n_rows = obj.cumulative_length.nda[start_row + n_rows - 1] - fd_start
            self.write_object(
                obj.flattened_data,
                "flattened_data",
                lh5_file,
                group=group,
                start_row=fd_start,
                n_rows=fd_n_rows,
                wo_mode=wo_mode,
                write_start=offset,
                hdf5_compression=hdf5_compression,
            )

            # now offset is used to give appropriate in-file values for
            # cumulative_length. Need to adjust it for start_row
            if start_row > 0:
                offset -= obj.cumulative_length.nda[start_row - 1]

            # Add offset to obj.cumulative_length itself to avoid memory allocation.
            # Then subtract it off after writing! (otherwise it will be changed
            # upon return)
            cl_dtype = obj.cumulative_length.nda.dtype.type
            obj.cumulative_length.nda += cl_dtype(offset)

            self.write_object(
                obj.cumulative_length,
                "cumulative_length",
                lh5_file,
                group=group,
                start_row=start_row,
                n_rows=n_rows,
                wo_mode=wo_mode,
                write_start=write_start,
                hdf5_compression=hdf5_compression,
            )
            obj.cumulative_length.nda -= cl_dtype(offset)

            return

        # if we get this far, must be one of the Array types
        elif isinstance(obj, Array):
            if n_rows is None or n_rows > obj.nda.shape[0] - start_row:
                n_rows = obj.nda.shape[0] - start_row

            nda = obj.nda[start_row : start_row + n_rows]

            # hack to store bools as uint8 for c / Julia compliance
            if nda.dtype.name == "bool":
                nda = nda.astype(np.uint8)

            # need to create dataset from ndarray the first time for speed
            # creating an empty dataset and appending to that is super slow!
            if (wo_mode != "a" and write_start == 0) or name not in group:
                maxshape = (None,) + nda.shape[1:]
                if wo_mode == "o" and name in group:
                    log.debug(f"overwriting {name} in {group}")
                    del group[name]

                # create HDF5 dataset
                # - compress using the 'compression' LGDO attribute, if
                #   available
                # - otherwise use "hdf5_compression"
                # - attach HDF5 dataset attributes, but not "compression"!
                comp_algo = obj.attrs.get("compression", hdf5_compression)
                comp_kwargs = {}
                if isinstance(comp_algo, str):
                    comp_kwargs = {"compression": comp_algo}
                elif comp_algo is not None:
                    comp_kwargs = comp_algo

                ds = group.create_dataset(
                    name, data=nda, maxshape=maxshape, **comp_kwargs
                )

                _attrs = obj.getattrs(datatype=True)
                _attrs.pop("compression", None)
                ds.attrs.update(_attrs)
                return

            # Now append or overwrite
            ds = group[name]
            if not isinstance(ds, h5py.Dataset):
                raise RuntimeError(
                    f"existing HDF5 object '{name}' in group '{group}'"
                    " is not a dataset! Cannot overwrite or append"
                )

            old_len = ds.shape[0]
            if wo_mode == "a":
                write_start = old_len
            add_len = write_start + nda.shape[0] - old_len
            ds.resize(old_len + add_len, axis=0)
            ds[write_start:] = nda
            return

        else:
            raise RuntimeError(
                f"do not know how to write '{name}' of type '{type(obj).__name__}'"
            )

    def read_n_rows(self, name: str, lh5_file: str | h5py.File) -> int | None:
        """Look up the number of rows in an Array-like object called `name` in
        `lh5_file`.

        Return ``None`` if it is a :class:`.Scalar` or a :class:`.Struct`."""
        # this is basically a stripped down version of read_object
        h5f = self.gimme_file(lh5_file, "r")
        if not h5f or name not in h5f:
            raise KeyError(f"'{name}' not in {lh5_file}")

        # get the datatype
        if "datatype" not in h5f[name].attrs:
            raise RuntimeError(
                f"'{name}' in file {lh5_file} is missing the datatype attribute"
            )

        datatype = h5f[name].attrs["datatype"]
        datatype, shape, elements = parse_datatype(datatype)

        # scalars are dim-0 datasets
        if datatype == "scalar":
            return None

        # structs don't have rows
        if datatype == "struct":
            return None

        # tables should have elements with all the same length
        if datatype == "table":
            # read out each of the fields
            rows_read = None
            for field in elements:
                n_rows_read = self.read_n_rows(name + "/" + field, h5f)
                if not rows_read:
                    rows_read = n_rows_read
                elif rows_read != n_rows_read:
                    log.warning(
                        f"'{field}' field in table '{name}' has {rows_read} rows, "
                        f"{n_rows_read} was expected"
                    )
            return rows_read

        # length of vector of vectors is the length of its cumulative_length
        if elements.startswith("array"):
            return self.read_n_rows(f"{name}/cumulative_length", h5f)

        # length of vector of encoded vectors is the length of its decoded_size
        if (
            elements.startswith("encoded_array")
            or datatype == "array_of_encoded_equalsized_arrays"
        ):
            return self.read_n_rows(f"{name}/encoded_data", h5f)

        # return array length (without reading the array!)
        if "array" in datatype:
            # compute the number of rows to read
            return h5f[name].shape[0]

        raise RuntimeError(f"don't know how to read datatype '{datatype}'")


def ls(lh5_file: str | h5py.Group, lh5_group: str = "") -> list[str]:
    """Return a list of LH5 groups in the input file and group, similar
    to ``ls`` or ``h5ls``. Supports wildcards in group names.


    Parameters
    ----------
    lh5_file
        name of file.
    lh5_group
        group to search. add a ``/`` to the end of the group name if you want to
        list all objects inside that group.
    """

    log.debug(
        f"Listing objects in '{lh5_file}'"
        + ("" if lh5_group == "" else f" (and group {lh5_group})")
    )

    lh5_st = LH5Store()
    # To use recursively, make lh5_file a h5group instead of a string
    if isinstance(lh5_file, str):
        lh5_file = lh5_st.gimme_file(lh5_file, "r")
        if lh5_group.startswith("/"):
            lh5_group = lh5_group[1:]

    if lh5_group == "":
        lh5_group = "*"

    splitpath = lh5_group.split("/", 1)
    matchingkeys = fnmatch.filter(lh5_file.keys(), splitpath[0])

    if len(splitpath) == 1:
        return matchingkeys
    else:
        ret = []
        for key in matchingkeys:
            ret.extend([f"{key}/{path}" for path in ls(lh5_file[key], splitpath[1])])
        return ret


def show(
    lh5_file: str | h5py.Group,
    lh5_group: str = "/",
    attrs: bool = False,
    indent: str = "",
    header: bool = True,
) -> None:
    """Print a tree of LH5 file contents with LGDO datatype.

    Parameters
    ----------
    lh5_file
        the LH5 file.
    lh5_group
        print only contents of this HDF5 group.
    attrs
        print the HDF5 attributes too.
    indent
        indent the diagram with this string.
    header
        print `lh5_group` at the top of the diagram.

    Examples
    --------
    >>> from pygama.lgdo import show
    >>> show("file.lh5", "/geds/raw")
    /geds/raw
    ├── channel · array<1>{real}
    ├── energy · array<1>{real}
    ├── timestamp · array<1>{real}
    ├── waveform · table{t0,dt,values}
    │   ├── dt · array<1>{real}
    │   ├── t0 · array<1>{real}
    │   └── values · array_of_equalsized_arrays<1,1>{real}
    └── wf_std · array<1>{real}
    """
    # open file
    if isinstance(lh5_file, str):
        lh5_file = h5py.File(expand_path(lh5_file), "r")

    # go to group
    if lh5_group != "/":
        lh5_file = lh5_file[lh5_group]

    if header:
        print(f"\033[1m{lh5_group}\033[0m")  # noqa: T201

    # get an iterator over the keys in the group
    it = iter(lh5_file)
    key = None

    # make sure there is actually something in this file/group
    try:
        key = next(it)  # get first key
    except StopIteration:
        print(f"{indent}└──  empty")  # noqa: T201
        return

    # loop over keys
    while True:
        val = lh5_file[key]
        # we want to print the LGDO datatype
        dtype = val.attrs.get("datatype", default="no datatype")
        if dtype == "no datatype" and isinstance(val, h5py.Group):
            dtype = "HDF5 group"

        attrs_d = dict(val.attrs)
        attrs_d.pop("datatype", "")
        attrs = "── " + str(attrs_d) if attrs_d else ""

        # is this the last key?
        killme = False
        try:
            k_new = next(it)  # get next key
        except StopIteration:
            char = "└──"
            killme = True  # we'll have to kill this loop later
        else:
            char = "├──"

        print(f"{indent}{char} \033[1m{key}\033[0m · {dtype} {attrs}")  # noqa: T201

        # if it's a group, call this function recursively
        if isinstance(val, h5py.Group):
            show(val, indent=indent + ("    " if killme else "│   "), header=False)

        # break or move to next key
        if killme:
            break
        else:
            key = k_new


def load_nda(
    f_list: str | list[str],
    par_list: list[str],
    lh5_group: str = "",
    idx_list: list[np.ndarray | list | tuple] = None,
) -> dict[str, np.ndarray]:
    r"""Build a dictionary of :class:`numpy.ndarray`\ s from LH5 data.

    Given a list of files, a list of LH5 table parameters, and an optional
    group path, return a NumPy array with all values for each parameter.

    Parameters
    ----------
    f_list
        A list of files. Can contain wildcards.
    par_list
        A list of parameters to read from each file.
    lh5_group
        group path within which to find the specified parameters.
    idx_list
        for fancy-indexed reads. Must be one index array for each file in
        `f_list`.

    Returns
    -------
    par_data
        A dictionary of the parameter data keyed by the elements of `par_list`.
        Each entry contains the data for the specified parameter concatenated
        over all files in `f_list`.
    """
    if isinstance(f_list, str):
        f_list = [f_list]
        if idx_list is not None:
            idx_list = [idx_list]
    if idx_list is not None and len(f_list) != len(idx_list):
        raise ValueError(
            f"f_list length ({len(f_list)}) != idx_list length ({len(idx_list)})!"
        )

    # Expand wildcards
    f_list = [f for f_wc in f_list for f in sorted(glob.glob(os.path.expandvars(f_wc)))]

    sto = LH5Store()
    par_data = {par: [] for par in par_list}
    for ii, f in enumerate(f_list):
        f = sto.gimme_file(f, "r")
        for par in par_list:
            if f"{lh5_group}/{par}" not in f:
                raise RuntimeError(f"'{lh5_group}/{par}' not in file {f_list[ii]}")

            if idx_list is None:
                data, _ = sto.read_object(f"{lh5_group}/{par}", f)
            else:
                data, _ = sto.read_object(f"{lh5_group}/{par}", f, idx=idx_list[ii])
            if not data:
                continue
            par_data[par].append(data.nda)
    par_data = {par: np.concatenate(par_data[par]) for par in par_list}
    return par_data


def load_dfs(
    f_list: str | list[str],
    par_list: list[str],
    lh5_group: str = "",
    idx_list: list[np.ndarray | list | tuple] = None,
) -> pd.DataFrame:
    """Build a :class:`pandas.DataFrame` from LH5 data.

    Given a list of files (can use wildcards), a list of LH5 columns, and
    optionally the group path, return a :class:`pandas.DataFrame` with all
    values for each parameter.

    See Also
    --------
    :func:`load_nda`

    Returns
    -------
    dataframe
        contains columns for each parameter in `par_list`, and rows containing
        all data for the associated parameters concatenated over all files in
        `f_list`.
    """
    return pd.DataFrame(
        load_nda(f_list, par_list, lh5_group=lh5_group, idx_list=idx_list)
    )


class LH5Iterator(Iterator):
    """
    A class for iterating through one or more LH5 files, one block of entries
    at a time. This also accepts an entry list/mask to enable event selection,
    and a field mask.

    This class can be used either for random access:

    >>> lh5_obj, n_rows = lh5_it.read(entry)

    to read the block of entries starting at entry. In case of multiple files
    or the use of an event selection, entry refers to a global event index
    across files and does not count events that are excluded by the selection.

    This can also be used as an iterator:

    >>> for lh5_obj, entry, n_rows in LH5Iterator(...):
    >>>    # do the thing!

    This is intended for if you are reading a large quantity of data but
    want to limit your memory usage (particularly when reading in waveforms!).
    The ``lh5_obj`` that is read by this class is reused in order to avoid
    reallocation of memory; this means that if you want to hold on to data
    between reads, you will have to copy it somewhere!
    """

    def __init__(
        self,
        lh5_files: str | list[str],
        groups: str | list[str],
        base_path: str = "",
        entry_list: list[int] | list[list[int]] = None,
        entry_mask: list[bool] | list[list[bool]] = None,
        field_mask: dict[str, bool] | list[str] | tuple[str] = None,
        buffer_len: int = 3200,
        friend: LH5Iterator = None,
    ) -> None:
        """
        Parameters
        ----------
        lh5_files
            file or files to read from. May include wildcards and environment
            variables.
        groups
            HDF5 group(s) to read. If a list is provided for both lh5_files
            and group, they must be the same size. If a file is wild-carded,
            the same group will be assigned to each file found
        entry_list
            list of entry numbers to read. If a nested list is provided,
            expect one top-level list for each file, containing a list of
            local entries. If a list of ints is provided, use global entries.
        entry_mask
            mask of entries to read. If a list of arrays is provided, expect
            one for each file. Ignore if a selection list is provided.
        field_mask
            mask of which fields to read. See :meth:`LH5Store.read_object` for
            more details.
        buffer_len
            number of entries to read at a time while iterating through files.
        friend
            a ''friend'' LH5Iterator that will be read in parallel with this.
            The friend should have the same length and entry list. A single
            LH5 table containing columns from both iterators will be returned.
        """
        self.lh5_st = LH5Store(base_path=base_path, keep_open=True)

        # List of files, with wildcards and env vars expanded
        if isinstance(lh5_files, str):
            lh5_files = [lh5_files]
            if isinstance(groups, list):
                lh5_files *= len(groups)
        elif not isinstance(lh5_files, list):
            raise ValueError("lh5_files must be a string or list of strings")

        if isinstance(groups, str):
            groups = [groups] * len(lh5_files)
        elif not isinstance(groups, list):
            raise ValueError("group must be a string or list of strings")

        if not len(groups) == len(lh5_files):
            raise ValueError("lh5_files and groups must have same length")

        self.lh5_files = []
        self.groups = []
        for f, g in zip(lh5_files, groups):
            f_exp = expand_path(f, list=True, base_path=base_path)
            self.lh5_files += f_exp
            self.groups += [g] * len(f_exp)

        if entry_list is not None and entry_mask is not None:
            raise ValueError(
                "entry_list and entry_mask arguments are mutually exclusive"
            )

        # Map to last row in each file
        self.file_map = np.full(len(self.lh5_files), np.iinfo("i").max, "i")
        # Map to last iterator entry for each file
        self.entry_map = np.full(len(self.lh5_files), np.iinfo("i").max, "i")
        self.buffer_len = buffer_len

        if len(self.lh5_files) > 0:
            f = self.lh5_files[0]
            g = self.groups[0]
            self.lh5_buffer = self.lh5_st.get_buffer(
                g,
                f,
                size=self.buffer_len,
                field_mask=field_mask,
            )
            self.file_map[0] = self.lh5_st.read_n_rows(g, f)
        else:
            raise RuntimeError(f"can't open any files from {lh5_files}")

        self.n_rows = 0
        self.current_entry = 0
        self.next_entry = 0

        self.field_mask = field_mask

        # List of entry indices from each file
        self.local_entry_list = None
        self.global_entry_list = None
        if entry_list is not None:
            entry_list = list(entry_list)
            if isinstance(entry_list[0], int):
                self.local_entry_list = [None] * len(self.file_map)
                self.global_entry_list = np.array(entry_list, "i")
                self.global_entry_list.sort()

            else:
                self.local_entry_list = [[]] * len(self.file_map)
                for i_file, local_list in enumerate(entry_list):
                    self.local_entry_list[i_file] = np.array(local_list, "i")
                    self.local_entry_list[i_file].sort()

        elif entry_mask is not None:
            # Convert entry mask into an entry list
            if isinstance(entry_mask, pd.Series):
                entry_mask = entry_mask.values
            if isinstance(entry_mask, np.ndarray):
                self.local_entry_list = [None] * len(self.file_map)
                self.global_entry_list = np.nonzero(entry_mask)[0]
            else:
                self.local_entry_list = [[]] * len(self.file_map)
                for i_file, local_mask in enumerate(entry_mask):
                    self.local_entry_list[i_file] = np.nonzero(local_mask)[0]

        # Attach the friend
        if friend is not None:
            if not isinstance(friend, LH5Iterator):
                raise ValueError("Friend must be an LH5Iterator")
            self.lh5_buffer.join(friend.lh5_buffer)
        self.friend = friend

    def _get_file_cumlen(self, i_file: int) -> int:
        """Helper to get cumulative file length of file"""
        if i_file < 0:
            return 0
        fcl = self.file_map[i_file]
        if fcl == np.iinfo("i").max:
            fcl = self._get_file_cumlen(i_file - 1) + self.lh5_st.read_n_rows(
                self.groups[i_file], self.lh5_files[i_file]
            )
            self.file_map[i_file] = fcl
        return fcl

    def _get_file_cumentries(self, i_file: int) -> int:
        """Helper to get cumulative iterator entries in file"""
        if i_file < 0:
            return 0
        n = self.entry_map[i_file]
        if n == np.iinfo("i").max:
            elist = self.get_file_entrylist(i_file)
            fcl = self._get_file_cumlen(i_file)
            if elist is None:
                # no entry list provided
                n = fcl
            else:
                file_entries = self.get_file_entrylist(i_file)
                # check that file entries fall inside of file
                if file_entries[-1] >= fcl:
                    logging.warning(f"Found entries out of range for file {i_file}")
                    n = np.searchsorted(file_entries, fcl, "right")
                else:
                    n = len(file_entries)
                n += self._get_file_cumentries(i_file - 1)
            self.entry_map[i_file] = n
        return n

    def get_file_entrylist(self, i_file: int) -> np.ndarray:
        """Helper to get entry list for file"""
        # If no entry list is provided
        if self.local_entry_list is None:
            return None

        elist = self.local_entry_list[i_file]
        if elist is None:
            # Get local entrylist for this file from global entry list
            f_start = self._get_file_cumlen(i_file - 1)
            f_end = self._get_file_cumlen(i_file)
            i_start = self._get_file_cumentries(i_file - 1)
            i_stop = np.searchsorted(self.global_entry_list, f_end, "right")
            elist = np.array(self.global_entry_list[i_start:i_stop], "i") - f_start
            self.local_entry_list[i_file] = elist
        return elist

    def get_global_entrylist(self) -> np.ndarray:
        """Get global entry list, constructing it if needed"""
        if self.global_entry_list is None and self.local_entry_list is not None:
            self.global_entry_list = np.zeros(len(self), "i")
            for i_file in range(len(self.lh5_files)):
                i_start = self.get_file_cumentries(i_file - 1)
                i_stop = self.get_file_cumentries(i_file)
                f_start = self.get_file_cumlen(i_file - 1)
                self.global_entry_list[i_start:i_stop] = (
                    self.get_file_entrylist(i_file) + f_start
                )
        return self.global_entry_list

    def read(self, entry: int) -> tuple[LGDO, int]:
        """Read the nextlocal chunk of events, starting at entry. Return the
        LH5 buffer and number of rows read."""
        self.n_rows = 0
        i_file = np.searchsorted(self.entry_map, entry, "right")

        # if file hasn't been opened yet, search through files
        # sequentially until we find the right one
        if i_file < len(self.lh5_files) and self.entry_map[i_file] == np.iinfo("i").max:
            while i_file < len(self.lh5_files) and entry >= self._get_file_cumentries(
                i_file
            ):
                i_file += 1

        if i_file == len(self.lh5_files):
            return (self.lh5_buffer, self.n_rows)
        local_entry = entry - self._get_file_cumentries(i_file - 1)

        while self.n_rows < self.buffer_len and i_file < len(self.file_map):
            # Loop through files
            local_idx = self.get_file_entrylist(i_file)
            i_local = local_idx[local_entry] if local_idx is not None else local_entry
            self.lh5_buffer, n_rows = self.lh5_st.read_object(
                self.groups[i_file],
                self.lh5_files[i_file],
                start_row=i_local,
                n_rows=self.buffer_len - self.n_rows,
                idx=local_idx,
                field_mask=self.field_mask,
                obj_buf=self.lh5_buffer,
                obj_buf_start=self.n_rows,
            )

            self.n_rows += n_rows
            i_file += 1
            local_entry = 0

        self.current_entry = entry

        if self.friend is not None:
            self.friend.read(entry)

        return (self.lh5_buffer, self.n_rows)

    def reset_field_mask(self, mask):
        """Replaces the field mask of this iterator and any friends with mask"""
        self.field_mask = mask
        if self.friend is not None:
            self.friend.reset_field_mask(mask)

    def __len__(self) -> int:
        """Return the total number of entries."""
        return (
            self._get_file_cumentries(len(self.lh5_files) - 1)
            if len(self.entry_map) > 0
            else 0
        )

    def __iter__(self) -> Iterator:
        """Loop through entries in blocks of size buffer_len."""
        self.current_entry = 0
        self.next_entry = 0
        return self

    def __next__(self) -> tuple[LGDO, int, int]:
        """Read next buffer_len entries and return lh5_table, iterator entry
        and n_rows read."""
        buf, n_rows = self.read(self.next_entry)
        self.next_entry = self.current_entry + n_rows
        if n_rows == 0:
            raise StopIteration
        return (buf, self.current_entry, n_rows)


@nb.njit(parallel=False, fastmath=True)
def _make_fd_idx(starts, stops, idx):
    k = 0
    if len(starts) < len(stops):
        for i in range(stops[0]):
            idx[k] = i
            k += 1
        stops = stops[1:]
    for j in range(len(starts)):
        for i in range(starts[j], stops[j]):
            idx[k] = i
            k += 1
    return (idx,)



r"""
Contains a list of DSP processors, implemented using Numba's
:func:`numba.guvectorize` to implement NumPy's :class:`numpy.ufunc` interface.
In other words, all of the functions are void functions whose outputs are given
as parameters.  The :class:`~numpy.ufunc` interface provides additional
information about the function signatures that enables broadcasting the arrays
and SIMD processing. Thanks to the :class:`~numpy.ufunc` interface, they can
also be called to return a NumPy array, but if this is done, memory will be
allocated for this array, slowing things down.

The pygama processors use the :class:`~numpy.ufunc` framework, which is
designed to encourage highly performant python practices. These functions have
several advantages:

1. They work with :class:`numpy.array`. NumPy arrays are arrays of same-typed
   objects that are stored adjacently in memory (equivalent to a dynamically
   allocated C-array or a C++ vector). Compared to standard Python lists, they
   perform computations much faster. They are ideal for representing waveforms
   and transformed waveforms.

2. They perform vectorized operations. Vectorization causes commands to
   perform the same operation on all components of an array. For example, the
   :class:`~numpy.ufunc` ``np.add(a, b, out=c)`` (equivalently ``c=a+b``) is
   equivalent to: ::

       for i in range(len(c)): c[i] = a[i] + b[i]

   Loops are slow in python since it is an interpreted language; vectorized
   commands remove the loop and only call the Python interpreter once.

   Furthermore, :class:`~numpy.ufunc`\ s are capable of `broadcasting
   <https://docs.scipy.org/doc/numpy/reference/ufuncs.html#broadcasting>`_
   their dimensions. This involves a safety check to ensure the dimensions of
   ``a`` and ``b`` are compatible sizes. It also will automatically replicate a
   ``size-1`` dimension over another one, enabling the addition of a scalar to
   a vector quantity. This is useful, as it allows us to process multiple
   waveforms at once.

   One of the biggest advantages of vectorized :class:`~numpy.ufunc`\ s is that
   many of them will take advantage of SIMD (same input-multiple data)
   vectorization on a vector-CPU. Modern CPUs typically have 256- or 512-bit
   wide processing units, which can accommodate multiple 32- or 64-bit numbers.
   Programming with these, however, is quite difficult and requires specialized
   commands to be called.  Luckily for us, many NumPy :class:`~numpy.ufunc`\ s
   will automatically use these for us, speeding up our code!

3. :class:`~numpy.ufunc`\ s are capable of calculating their output in place,
   meaning they can place calculated values in pre-allocated memory rather than
   allocating and returning new values. This is important because memory
   allocation is one of the slowest processes computers can perform, and should
   be avoided. With :class:`~numpy.ufunc`\ s, this can be done using the out
   keyword in arguments (ex ``numpy.add(a, b, out=c)``, or more succinctly,
   ``numpy.add(a, b, c)``).  While this may seem counterintuitive at first, the
   alternative (``c = np.add(a,b)`` or ``c = a+b``) causes an entirely new
   array to be allocated, with c pointing at that. These array allocations can
   add up very quickly: ``e = a*b + c*d``, for example, would allocate 3
   different arrays: one for ``a*b``, one for ``c*d``, and one for the sum of
   those two. As we write :class:`~numpy.ufunc`\ s, it is important that we try
   to use functions that operate in place as much as possible!
"""

from .bl_subtract import bl_subtract
from .convolutions import cusp_filter, t0_filter, zac_filter
from .dplms import dplms
from .dwt import discrete_wavelet_transform
from .fftw import dft, inv_dft, psd
from .fixed_time_pickoff import fixed_time_pickoff
from .gaussian_filter1d import gaussian_filter1d
from .get_multi_local_extrema import get_multi_local_extrema
from .histogram import histogram, histogram_stats
from .linear_slope_fit import linear_slope_diff, linear_slope_fit
from .log_check import log_check
from .min_max import min_max
from .moving_windows import (
    avg_current,
    moving_window_left,
    moving_window_multi,
    moving_window_right,
)
from .multi_a_filter import multi_a_filter
from .multi_t_filter import multi_t_filter, remove_duplicates
from .optimize import optimize_1pz, optimize_2pz
from .param_lookup import param_lookup
from .peak_snr_threshold import peak_snr_threshold
from .pole_zero import double_pole_zero, pole_zero
from .presum import presum
from .pulse_injector import inject_exp_pulse, inject_sig_pulse
from .saturation import saturation
from .soft_pileup_corr import soft_pileup_corr, soft_pileup_corr_bl
from .time_over_threshold import time_over_threshold
from .time_point_thresh import interpolated_time_point_thresh, time_point_thresh
from .trap_filters import asym_trap_filter, trap_filter, trap_norm, trap_pickoff
from .upsampler import interpolating_upsampler, upsampler
from .wiener_filter import wiener_filter
from .windower import windower

__all__ = [
    "bl_subtract",
    "cusp_filter",
    "t0_filter",
    "zac_filter",
    "discrete_wavelet_transform",
    "dft",
    "inv_dft",
    "psd",
    "fixed_time_pickoff",
    "gaussian_filter1d",
    "get_multi_local_extrema",
    "histogram",
    "histogram_stats",
    "linear_slope_fit",
    "linear_slope_diff",
    "log_check",
    "min_max",
    "avg_current",
    "moving_window_left",
    "moving_window_multi",
    "moving_window_right",
    "multi_a_filter",
    "multi_t_filter",
    "remove_duplicates",
    "optimize_1pz",
    "optimize_2pz",
    "param_lookup",
    "double_pole_zero",
    "pole_zero",
    "presum",
    "inject_exp_pulse",
    "inject_sig_pulse",
    "saturation",
    "peak_snr_threshold",
    "soft_pileup_corr",
    "soft_pileup_corr_bl",
    "time_point_thresh",
    "interpolated_time_point_thresh",
    "asym_trap_filter",
    "trap_filter",
    "trap_norm",
    "trap_pickoff",
    "upsampler",
    "interpolating_upsampler",
    "wiener_filter",
    "windower",
    "time_over_threshold",
    "dplms",
]



import pytest

import pygama.lgdo as lgdo


def test_datatype_name():
    scalar = lgdo.Scalar(value=42)
    assert scalar.datatype_name() == "real"


def test_form_datatype():
    scalar = lgdo.Scalar(value=42)
    assert scalar.form_datatype() == "real"


def test_init():
    attrs = {"attr1": 1}
    scalar = lgdo.Scalar(value=42, attrs=attrs)
    assert scalar.value == 42
    assert scalar.attrs == attrs | {"datatype": "real"}

    with pytest.raises(ValueError):
        lgdo.Scalar(value=42, attrs={"datatype": "string"})


def test_getattrs():
    scalar = lgdo.Scalar(value=42, attrs={"attr1": 1})
    assert scalar.getattrs() == {"attr1": 1}
    assert scalar.getattrs(True) == {"attr1": 1, "datatype": "real"}


def test_equality():
    assert lgdo.Scalar(value=42) == lgdo.Scalar(value=42)



from __future__ import annotations

import logging

import numpy as np

from pygama.raw.compass.compass_event_decoder import CompassEventDecoder
from pygama.raw.compass.compass_header_decoder import CompassHeaderDecoder
from pygama.raw.data_decoder import DataDecoder
from pygama.raw.data_streamer import DataStreamer
from pygama.raw.raw_buffer import RawBuffer, RawBufferLibrary

log = logging.getLogger(__name__)


class CompassStreamer(DataStreamer):
    """Data streamer for CoMPASS data streams."""

    def __init__(self, compass_config_file=None) -> None:
        super().__init__()
        self.in_stream = None
        self.header = None
        self.buffer = np.empty(
            1024, dtype="bytes"
        )  # create a buffer that is around 4 kB
        self.header_decoder = CompassHeaderDecoder()
        self.decoder_id_dict = {}  # dict of data_id to decoder object
        self.decoder_name_dict = {}  # dict of name to decoder object
        self.compass_config_file = compass_config_file  # optional config file to be passed when calling build_raw
        self.event_rbkd = None

    def get_decoder_list(self) -> list[DataDecoder]:
        dec_list = []
        dec_list.append(self.header_decoder)
        dec_list.append(self.event_decoder)
        return list(dec_list)

    def open_stream(
        self,
        stream_name: str,
        rb_lib: RawBufferLibrary = None,
        buffer_size: int = 8192,
        chunk_mode: str = "any_full",
        out_stream: str = "",
    ) -> tuple[list[RawBuffer], int]:
        """Initialize the CoMPASS data stream. If a config file is present, load just the 2 byte header.
        If a config file is absent, read the first packet to determine waveform length, and initialize the
        keys to the max number of channels of a CAEN digitizer.

        Parameters
        ----------
        stream_name
            The CoMPASS filename. Only file streams are currently supported.
            Socket stream reading can be added later.
        rb_lib
            library of buffers for this stream.
        buffer_size
            length of tables to be read out in :meth:`read_chunk`.
        chunk_mode : 'any_full', 'only_full', or 'single_packet'
            sets the mode use for :meth:`read_chunk`.
        out_stream
            optional name of output stream for default `rb_lib` generation.

        Returns
        -------
        header_data
            a list of length 1 containing the raw buffer holding the CoMPASS header.

        Notes
        -----
        CoMPASS files have a header that is only 2 bytes long and only appears at the top of the file.
        So, we must read this header once, and then proceed to read packets in.
        """
        # If a config file is not present, the wf_len can be determined by opening the first few bytes of the in_stream
        wf_len = None
        if self.compass_config_file is None:
            self.set_in_stream(stream_name)

            first_bytes = self.in_stream.read(27)

            energy_short = str(
                bin(int.from_bytes(first_bytes[:2], byteorder="little"))
            )[::-1][2]

            if int(energy_short) == 1:
                [wf_len] = np.frombuffer(first_bytes[23:27], dtype=np.uint32)
            else:
                [wf_len] = np.frombuffer(first_bytes[21:25], dtype=np.uint32)

            self.close_stream()

        # set the in_stream
        self.set_in_stream(stream_name)
        self.n_bytes_read = 0

        # read in and decode the file header info, passing the compass_config_file, if present
        self.header = self.header_decoder.decode_header(
            self.in_stream, self.compass_config_file, wf_len
        )  # returns an lgdo.Struct
        self.n_bytes_read += (
            2  # there are 2 bytes in the header, for a 16 bit number to read out
        )

        # set up data loop variables
        self.packet_id = 0

        # instantiate the event decoder and pass the header
        self.event_decoder = CompassEventDecoder(self.header)

        # initialize the buffers in rb_lib. Store them for fast lookup
        super().open_stream(
            stream_name,
            rb_lib,
            buffer_size=buffer_size,
            chunk_mode=chunk_mode,
            out_stream=out_stream,
        )
        if rb_lib is None:
            rb_lib = self.rb_lib

        self.event_rbkd = (
            rb_lib["CompassEventDecoder"].get_keyed_dict()
            if "CompassEventDecoder" in rb_lib
            else None
        )

        # set up the header raw buffer
        if "CompassHeaderDecoder" in rb_lib:
            header_rb_list = rb_lib["CompassHeaderDecoder"]
            if len(header_rb_list) != 1:
                log.warning(
                    f"header_rb_list had length {len(header_rb_list)}, ignoring all but the first"
                )
            rb = header_rb_list[0]
        else:
            rb = RawBuffer(lgdo=self.header_decoder.make_lgdo())
        rb.loc = 1  # we have filled this buffer
        return [rb]

    def set_in_stream(self, stream_name: str) -> None:
        """Sets the in_stream by opening a binary file"""
        if self.in_stream is not None:
            self.in_stream.close()
            self.in_stream = None
        else:
            self.in_stream = open(
                stream_name.encode("utf-8"), "rb"
            )  # CoMPASS files are binary
        self.n_bytes_read = 0

    def close_stream(self) -> None:
        """Close this data stream."""
        if self.in_stream is None:
            raise RuntimeError("tried to close an unopened stream")
        self.in_stream.close()
        self.in_stream = None

    def load_packet(self) -> np.uint32 | None:
        """Loads the next packet into the internal buffer.

        Returns packet as a :class:`numpy.uint32` view of the buffer (a slice),
        returns ``None`` at EOF.

        Notes
        -----
        First, load_packet finds the packet's metadata length in bytes by looking at the header,
        and then reads the metadata into the buffer. From the metadata, the length of the waveform can be determined and then the buffer is
        resized to fit the metadata and the waveform. Finally, the buffer is filled with the packet and then returned.

        """
        if self.in_stream is None:
            raise RuntimeError("self.in_stream is None")

        if (self.packet_id == 0) and (self.n_bytes_read != 2):
            raise RuntimeError(
                f"The 2 byte filer header was not converted, instead read in {self.n_bytes_read} for the file header"
            )

        # packets have metadata of variable lengths, depending on if the header shows that energy_short is present in the metadata
        if int(self.header["energy_short"].value) == 1:
            header_length = 25  # if the energy short is present, then there are an extra 2 bytes in the metadata
        else:
            header_length = 23  # the normal packet metadata is 23 bytes long

        # read the packet's metadata into the buffer
        pkt_hdr = self.buffer[:header_length]
        n_bytes_read = self.in_stream.readinto(pkt_hdr)
        self.n_bytes_read += n_bytes_read

        # return None once we run out of file
        if n_bytes_read == 0:
            return None
        if (n_bytes_read != 25) and (n_bytes_read != 23):
            raise RuntimeError(f"only got {n_bytes_read} bytes for packet header")

        # get the waveform length so we can read in the rest of the packet
        if n_bytes_read == 25:
            [num_samples] = np.frombuffer(pkt_hdr[21:25], dtype=np.uint32)
            pkt_length = header_length + 2 * num_samples
        if n_bytes_read == 23:
            [num_samples] = np.frombuffer(pkt_hdr[19:23], dtype=np.uint32)
            pkt_length = header_length + 2 * num_samples

        # load into buffer, resizing as necessary
        if len(self.buffer) < pkt_length:
            self.buffer.resize(pkt_length, refcheck=False)
        n_bytes_read = self.in_stream.readinto(self.buffer[header_length:pkt_length])
        self.n_bytes_read += n_bytes_read
        if n_bytes_read != pkt_length - header_length:
            raise RuntimeError(
                f"only got {n_bytes_read} bytes for packet read when {pkt_length-header_length} were expected."
            )

        # return just the packet
        return self.buffer[:pkt_length]

    def read_packet(self) -> bool:
        """Read and decode a packet of data.
        Data written to the `event_rbkd` attribute.
        """
        packet = self.load_packet()
        # Return True until we have no more file left to read
        if packet is None:
            return False
        self.packet_id += 1
        self.any_full |= self.event_decoder.decode_packet(
            packet, self.event_rbkd, self.packet_id, self.header
        )

        return True



import numpy as np

from pygama.lgdo import (
    Array,
    ArrayOfEncodedEqualSizedArrays,
    Scalar,
    VectorOfEncodedVectors,
    VectorOfVectors,
)


def test_voev_init():
    voev = VectorOfEncodedVectors(
        VectorOfVectors(shape_guess=(100, 1000), dtype="uint16")
    )
    assert len(voev.decoded_size) == 100
    assert voev.attrs["datatype"] == "array<1>{encoded_array<1>{real}}"
    assert len(voev) == 100

    voev = VectorOfEncodedVectors(
        encoded_data=VectorOfVectors(shape_guess=(100, 1000), dtype="uint16"),
        decoded_size=Array(shape=100),
        attrs={"sth": 1},
    )
    assert voev.attrs == {"datatype": "array<1>{encoded_array<1>{real}}", "sth": 1}


def test_aoeesa_init():
    voev = ArrayOfEncodedEqualSizedArrays(
        VectorOfVectors(shape_guess=(100, 1000), dtype="uint16")
    )
    assert isinstance(voev.decoded_size, Scalar)
    assert voev.attrs["datatype"] == "array_of_encoded_equalsized_arrays<1,1>{real}"
    assert len(voev) == 100

    voev = ArrayOfEncodedEqualSizedArrays(
        encoded_data=VectorOfVectors(shape_guess=(100, 1000), dtype="uint16"),
        decoded_size=99,
        attrs={"sth": 1},
    )
    assert voev.decoded_size.value == 99
    assert voev.attrs == {
        "datatype": "array_of_encoded_equalsized_arrays<1,1>{real}",
        "sth": 1,
    }


def test_resize():
    voev = VectorOfEncodedVectors(
        encoded_data=VectorOfVectors(shape_guess=(100, 1000), dtype="uint16"),
        decoded_size=Array(shape=100),
    )
    voev.resize(50)
    assert len(voev) == 50

    voev = ArrayOfEncodedEqualSizedArrays(
        encoded_data=VectorOfVectors(shape_guess=(100, 1000), dtype="uint16"),
        decoded_size=99,
    )
    voev.resize(50)
    assert len(voev) == 50


def test_append():
    voev = VectorOfEncodedVectors(
        encoded_data=VectorOfVectors(shape_guess=(100, 1000), dtype="uint16"),
        decoded_size=Array(shape=100),
    )
    voev.append(([1, 3, 5], 99))
    assert len(voev) == 101
    assert (voev[-1][0] == [1, 3, 5]).all()
    assert voev[-1][1] == 99

    voev = ArrayOfEncodedEqualSizedArrays(
        encoded_data=VectorOfVectors(shape_guess=(100, 1000), dtype="uint16"),
        decoded_size=100,
    )
    voev.append([1, 3, 5])
    assert len(voev) == 101
    assert (voev[-1] == [1, 3, 5]).all()
    assert voev.decoded_size.value == 100


def test_insert():
    voev = VectorOfEncodedVectors(
        encoded_data=VectorOfVectors(
            shape_guess=(100, 1000), dtype="uint16", fill_val=0
        ),
        decoded_size=Array(shape=100, fill_val=12),
    )
    voev.insert(3, ([1, 3, 5], 99))
    assert len(voev) == 101
    assert (voev[3][0] == [1, 3, 5]).all()
    assert voev[3][1] == 99

    voev = ArrayOfEncodedEqualSizedArrays(
        encoded_data=VectorOfVectors(
            shape_guess=(100, 1000), dtype="uint16", fill_val=0
        ),
        decoded_size=100,
    )

    voev.insert(4, [1, 3, 5])
    assert len(voev) == 101
    assert (voev[4] == [1, 3, 5]).all()
    assert voev.decoded_size.value == 100


def test_replace():
    voev = VectorOfEncodedVectors(
        encoded_data=VectorOfVectors(
            shape_guess=(100, 1000), dtype="uint16", fill_val=0
        ),
        decoded_size=Array(shape=100, fill_val=12),
    )
    voev.replace(3, ([1, 3, 5], 99))
    assert len(voev) == 100
    assert (voev[3][0] == [1, 3, 5]).all()
    assert voev[3][1] == 99

    voev = ArrayOfEncodedEqualSizedArrays(
        encoded_data=VectorOfVectors(
            shape_guess=(100, 1000), dtype="uint16", fill_val=0
        ),
        decoded_size=100,
    )

    voev.replace(4, [1, 3, 5])
    assert len(voev) == 100
    assert (voev[4] == [1, 3, 5]).all()
    assert voev.decoded_size.value == 100


def test_voev_set_get_vector():
    voev = VectorOfEncodedVectors(
        encoded_data=VectorOfVectors(shape_guess=(100, 3), dtype="uint16", fill_val=0),
        decoded_size=Array(shape=100),
        attrs={"sth": 1},
    )
    voev[5] = (np.array([1, 2, 3]), 7)
    assert np.array_equal(voev[5][0], np.array([1, 2, 3]))
    assert voev[5][1] == 7

    assert np.array_equal(voev[5][0], np.array([1, 2, 3]))
    assert voev[5][1] == 7

    voev[6] = (np.array([1, 2, 3]), 7)
    assert np.array_equal(voev[6][0], np.array([1, 2, 3]))
    assert voev[6][1] == 7


def test_aoeesa_set_get_vector():
    voev = ArrayOfEncodedEqualSizedArrays(
        encoded_data=VectorOfVectors(shape_guess=(100, 3), dtype="uint16", fill_val=0),
        decoded_size=99,
        attrs={"sth": 1},
    )
    voev[5] = np.array([1, 2, 3])
    assert np.array_equal(voev[5], np.array([1, 2, 3]))

    assert np.array_equal(voev[5], np.array([1, 2, 3]))

    voev[6] = np.array([1, 2, 3])
    assert np.array_equal(voev[6], np.array([1, 2, 3]))

    assert voev.decoded_size.value == 99


def test_voev_iteration():
    voev = VectorOfEncodedVectors(
        encoded_data=VectorOfVectors(
            flattened_data=Array(nda=np.array([1, 2, 3, 4, 5, 2, 4, 8, 9, 7, 5, 3, 1])),
            cumulative_length=Array(nda=np.array([2, 5, 6, 10, 13])),
        ),
        decoded_size=Array(shape=5, fill_val=6),
    )

    desired = [
        [1, 2],
        [3, 4, 5],
        [2],
        [4, 8, 9, 7],
        [5, 3, 1],
    ]

    for i, (v, s) in enumerate(voev):
        assert np.array_equal(v, desired[i])
        assert s == 6


def test_aoeesa_iteration():
    voev = ArrayOfEncodedEqualSizedArrays(
        encoded_data=VectorOfVectors(
            flattened_data=Array(nda=np.array([1, 2, 3, 4, 5, 2, 4, 8, 9, 7, 5, 3, 1])),
            cumulative_length=Array(nda=np.array([2, 5, 6, 10, 13])),
        ),
        decoded_size=99,
    )

    desired = [
        [1, 2],
        [3, 4, 5],
        [2],
        [4, 8, 9, 7],
        [5, 3, 1],
    ]

    for i, v in enumerate(voev):
        assert np.array_equal(v, desired[i])



from __future__ import annotations

import copy
import logging

import numpy as np

from pygama import lgdo
from pygama.raw.data_decoder import DataDecoder

log = logging.getLogger(__name__)


def get_bc(board: int, channel: int) -> int:
    """Create a standard hash for the board and channel of a CoMPASS file."""
    return (board << 9) + (channel & 0xF)


compass_decoded_values = {
    # packet index in file
    "packet_id": {"dtype": "uint32"},
    # ID of board
    "board": {"dtype": "uint32"},
    # ID of channel recording data
    "channel": {"dtype": "uint32"},
    # Timestamp of event
    "timestamp": {"dtype": "float64", "units": "ps"},
    # Energy of event
    "energy": {"dtype": "uint32"},
    # Energy short of event
    "energy_short": {"dtype": "uint32"},
    # Flags that the digitizer raised
    "flags": {"dtype": "uint32"},
    # number of samples in a waveform
    "num_samples": {"dtype": "int64"},
    # waveform data
    "waveform": {
        "dtype": "uint16",
        "datatype": "waveform",
        "wf_len": 65532,  # max value. override this before initializing buffers to save RAM
        "dt": 16,  # override if a different clock rate is used
        "dt_units": "ns",
        "t0_units": "ns",
    },
}
"""Default CoMPASS Event decoded values.

Warning
-------
This configuration can be dynamically modified by the decoder at runtime.
"""


class CompassEventDecoder(DataDecoder):
    """
    Decode CAEN digitizer event data.
    """

    def __init__(self, header=None, *args, **kwargs) -> None:
        self.decoded_values = {}
        super().__init__(*args, **kwargs)
        self.skipped_channels = {}
        if header is not None:
            self.set_header(header)

    # set the decoded_values for each board/channel combination because they could have different settings
    def set_header(self, header):
        self.header = header

        # Loop over crates, cards, build decoded values for enabled channels
        for board in self.header["boards"].keys():
            for channel in self.header["boards"][board]["channels"].keys():
                bc = get_bc(int(board), int(channel))

                self.decoded_values[bc] = copy.deepcopy(compass_decoded_values)

                # get trace length(s). Should all be the same
                self.decoded_values[bc]["waveform"]["wf_len"] = int(
                    float(
                        self.header["boards"][board]["wf_len"].value
                    )  # the header is a struct, so we need to return its value
                )

    def get_key_lists(self) -> list[list[str]]:
        key_lists = []
        for key in self.decoded_values.keys():
            key_lists.append(key)
        return [key_lists]

    def get_decoded_values(self, key=None):
        if key is None:
            dec_vals_list = self.decoded_values.values()
            if len(dec_vals_list) >= 0:
                return list(dec_vals_list)[0]
            raise RuntimeError("decoded_values not built")
        if key in self.decoded_values:
            return self.decoded_values[key]
        raise KeyError(f"no decoded values for key {key}")

    def decode_packet(
        self,
        packet: bytes,
        evt_rbkd: lgdo.Table | dict[int, lgdo.Table],
        packet_id: int,
        header: lgdo.Table | dict[int, lgdo.Table],
    ) -> bool:
        """Access ``CoMPASSEvent`` members for each event in the DAQ file.

        Parameters
        ----------
        packet
            The packet to be decoded
        evt_rbkd
            A single table for reading out all data, or a dictionary of tables
            keyed by channel number.
        packet_id
            The index of the packet in the `CoMPASS` stream. Incremented by
            :class:`~.raw.compass.compass_streamer.CompassStreamer`.
        header
            The header of the CoMPASS file, along with user config info,
            that is used to determine waveform lengths and thus buffer sizes

        Returns
        -------
        n_bytes
            (estimated) number of bytes in the packet that was just decoded.
        """
        # Read in the board number and channel number so we can get the right entry in the table
        # Regardless of the length of the header, the first 4 bytes are the board and the channel number
        board = np.frombuffer(packet[0:2], dtype=np.uint16)[0]
        channel = np.frombuffer(packet[2:4], dtype=np.uint16)[0]

        bc = get_bc(board, channel)

        # get the table for this channel
        if bc not in evt_rbkd:
            if bc not in self.skipped_channels:
                self.skipped_channels[bc] = 0
                log.debug(f"Skipping channel: {channel}")
                log.debug(f"evt_rbkd: {evt_rbkd.keys()}")
            self.skipped_channels[bc] += 1
            return False
        tbl = evt_rbkd[bc].lgdo
        ii = evt_rbkd[bc].loc

        # store packet id
        tbl["packet_id"].nda[ii] = packet_id

        # store the info we already have read in
        tbl["board"].nda[ii] = board
        tbl["channel"].nda[ii] = channel

        # the time stamp also does not care about if we have an energy short present
        tbl["timestamp"].nda[ii] = np.frombuffer(packet[4:12], dtype=np.uint64)[0]

        # get the rest of the values depending on if there is an energy_short present
        if int(header["energy_short"].value) == 1:  # again, the header is a struct
            tbl["energy"].nda[ii] = np.frombuffer(packet[12:14], dtype=np.uint16)[0]
            tbl["energy_short"].nda[ii] = np.frombuffer(packet[14:16], dtype=np.uint16)[
                0
            ]
            tbl["flags"].nda[ii] = np.frombuffer(packet[16:20], np.uint32)[0]
            tbl["num_samples"].nda[ii] = np.frombuffer(packet[21:25], dtype=np.uint32)[
                0
            ]

            if (
                tbl["num_samples"].nda[ii]
                != self.decoded_values[bc]["waveform"]["wf_len"]
            ):  # make sure that the waveform we read in is the same length as in the config
                raise RuntimeError(
                    f"Waveform size {tbl['num_samples'].nda[ii]} doesn't match expected size {self.decoded_values[bc]['waveform']['wf_len']}. "
                    "Skipping packet"
                )

            tbl["waveform"]["values"].nda[ii] = np.frombuffer(
                packet[25:], dtype=np.uint16
            )

        else:
            tbl["energy"].nda[ii] = np.frombuffer(packet[12:14], dtype=np.uint16)[0]
            tbl["energy_short"].nda[ii] = None
            tbl["flags"].nda[ii] = np.frombuffer(packet[14:18], np.uint32)[0]
            tbl["num_samples"].nda[ii] = np.frombuffer(packet[19:23], dtype=np.uint32)[
                0
            ]

            if (
                tbl["num_samples"].nda[ii]
                != self.decoded_values[bc]["waveform"]["wf_len"]
            ):  # make sure that the waveform we read in is the same length as in the config
                raise RuntimeError(
                    f"Waveform size {tbl['num_samples'].nda[ii]} doesn't match expected size {self.decoded_values[bc]['waveform']['wf_len']}. "
                    "Skipping packet"
                )

            tbl["waveform"]["values"].nda[ii] = np.frombuffer(
                packet[23:], dtype=np.uint16
            )

        evt_rbkd[bc].loc += 1
        return evt_rbkd[bc].is_full()



import os

import pygama.lgdo.lh5_store as store
from pygama.dsp import build_dsp


def test_histogram_fixed_width(lgnd_test_data):
    dsp_file = "/tmp/LDQTA_r117_20200110T105115Z_cal_geds__numpy_test_dsp.lh5"
    dsp_config = {
        "outputs": ["hist_weights", "hist_borders"],
        "processors": {
            "hist_weights , hist_borders": {
                "function": "histogram",
                "module": "pygama.dsp.processors.histogram",
                "args": ["waveform", "hist_weights(100)", "hist_borders(101)"],
                "unit": ["none", "ADC"],
            }
        },
    }
    build_dsp(
        f_raw=lgnd_test_data.get_path(
            "lh5/LDQTA_r117_20200110T105115Z_cal_geds_raw.lh5"
        ),
        f_dsp=dsp_file,
        dsp_config=dsp_config,
        write_mode="r",
    )
    assert os.path.exists(dsp_file)

    df = store.load_nda(dsp_file, ["hist_weights", "hist_borders"], "geds/dsp/")

    assert len(df["hist_weights"][0]) + 1 == len(df["hist_borders"][0])
    for i in range(2, len(df["hist_borders"][0])):
        a = df["hist_borders"][0][i - 1] - df["hist_borders"][0][i - 2]
        b = df["hist_borders"][0][i] - df["hist_borders"][0][i - 1]
        assert round(a, 2) == round(b, 2)



import scipy

from pygama.pargen import energy_optimisation  # noqa: F401


def test_import():
    pass


def test_scipy_version():
    assert scipy.__version__ != ""
    assert scipy.__version__ is not None



from __future__ import annotations

import gzip
import json
import logging

import numpy as np

from pygama.raw.data_streamer import DataStreamer
from pygama.raw.orca import orca_packet
from pygama.raw.orca.orca_base import OrcaDecoder
from pygama.raw.orca.orca_digitizers import (  # noqa: F401
    ORSIS3302DecoderForEnergy,
    ORSIS3316WaveformDecoder,
)
from pygama.raw.orca.orca_flashcam import (  # noqa: F401; ORFlashCamListenerStatusDecoder,
    ORFlashCamADCWaveformDecoder,
    ORFlashCamListenerConfigDecoder,
    ORFlashCamWaveformDecoder,
)
from pygama.raw.orca.orca_header_decoder import OrcaHeaderDecoder
from pygama.raw.orca.orca_run_decoder import ORRunDecoderForRun  # noqa: F401
from pygama.raw.raw_buffer import RawBuffer, RawBufferLibrary

log = logging.getLogger(__name__)


class OrcaStreamer(DataStreamer):
    """Data streamer for ORCA data."""

    def __init__(self) -> None:
        super().__init__()
        self.in_stream = None
        self.buffer = np.empty(1024, dtype="uint32")  # start with a 4 kB packet buffer
        self.header = None
        self.header_decoder = OrcaHeaderDecoder()
        self.decoder_id_dict = {}  # dict of data_id to decoder object
        self.rbl_id_dict = {}  # dict of RawBufferLists for each data_id

    # TODO: need to correct for endianness?
    def load_packet(self, skip_unknown_ids: bool = False) -> np.uint32 | None:
        """Loads the next packet into the internal buffer.

        Returns packet as a :class:`numpy.uint32` view of the buffer (a slice),
        returns ``None`` at EOF.
        """
        if self.in_stream is None:
            raise RuntimeError("self.in_stream is None")

        # read packet header
        pkt_hdr = self.buffer[:1]
        n_bytes_read = self.in_stream.readinto(pkt_hdr)  # buffer is at least 4 kB long
        self.n_bytes_read += n_bytes_read
        if n_bytes_read == 0:
            return None
        if n_bytes_read != 4:
            raise RuntimeError(f"only got {n_bytes_read} bytes for packet header")

        # if it's a short packet, we are done
        if orca_packet.is_short(pkt_hdr):
            return pkt_hdr

        # long packet: get length and check if we can skip it
        n_words = orca_packet.get_n_words(pkt_hdr)
        if (
            skip_unknown_ids
            and orca_packet.get_data_id(pkt_hdr, shift=False)
            not in self.decoder_id_dict
        ):
            self.in_stream.seek((n_words - 1) * 4, 1)
            self.n_bytes_read += (n_words - 1) * 4  # well, we didn't really read it...
            return pkt_hdr

        # load into buffer, resizing as necessary
        if len(self.buffer) < n_words:
            self.buffer.resize(n_words, refcheck=False)
        n_bytes_read = self.in_stream.readinto(self.buffer[1:n_words])
        self.n_bytes_read += n_bytes_read
        if n_bytes_read != (n_words - 1) * 4:
            log.error(
                f"only got {n_bytes_read} bytes for packet read when {(n_words-1)*4} were expected. Flushing all buffers and quitting..."
            )
            return None

        # return just the packet
        return self.buffer[:n_words]

    def get_decoder_list(self) -> list[OrcaDecoder]:
        return list(self.decoder_id_dict.values())

    def set_in_stream(self, stream_name: str) -> None:
        if self.in_stream is not None:
            self.close_in_stream()
        if stream_name.endswith(".gz"):
            self.in_stream = gzip.open(stream_name.encode("utf-8"), "rb")
        else:
            self.in_stream = open(stream_name.encode("utf-8"), "rb")
        self.n_bytes_read = 0

    def close_in_stream(self) -> None:
        if self.in_stream is None:
            raise RuntimeError("tried to close an unopened stream")
        self.in_stream.close()
        self.in_stream = None

    def close_stream(self) -> None:
        self.close_in_stream()

    def is_orca_stream(stream_name: str) -> bool:  # noqa: N805
        orca = OrcaStreamer()
        orca.set_in_stream(stream_name)
        first_bytes = orca.in_stream.read(12)
        orca.close_in_stream()

        # that read should have succeeded
        if len(first_bytes) != 12:
            log.debug(f"first 12B read only returned {first_bytes}B: not orca")
            return False

        # first 14 bits should be zero
        uints = np.frombuffer(first_bytes, dtype="uint32")
        if (uints[0] & 0xFFFC0000) != 0:
            log.debug(
                f"first fourteen bits non-zero ({uints[0] & 0xFFFC0000}): not orca"
            )
            return False

        # xml header length should fit within header packet length
        pad = uints[0] * 4 - 8 - uints[1]
        if pad < 0 or pad > 3:
            log.debug(
                f"header length = {uints[1]}B doesn't fit right within header packet length = {uints[0]*4-8}B: not orca"
            )
            return False

        # last 4 chars should be '<?xm'
        if first_bytes[8:].decode() != "<?xm":
            log.debug(
                f"last 4 chars of first 12 bytes = {first_bytes[8:].decode()} != '<?xm': not orca"
            )
            return False

        # it must be an orca stream
        return True

    def hex_dump(
        self,
        stream_name: str,
        n_packets: int = np.inf,
        skip_header: bool = False,
        shift_data_id: bool = True,
        print_n_words: bool = False,
        max_words: int = np.inf,
        as_int: bool = False,
        as_short: bool = False,
    ) -> None:
        self.set_in_stream(stream_name)
        if skip_header:
            self.load_packet()
        while n_packets > 0:
            packet = self.load_packet()
            if packet is None:
                self.close_in_stream()
                return
            orca_packet.hex_dump(
                packet,
                shift_data_id=shift_data_id,
                print_n_words=print_n_words,
                max_words=max_words,
                as_int=as_int,
                as_short=as_short,
            )
            n_packets -= 1

    def open_stream(
        self,
        stream_name: str,
        rb_lib: RawBufferLibrary = None,
        buffer_size: int = 8192,
        chunk_mode: str = "any_full",
        out_stream: str = "",
    ) -> list[RawBuffer]:
        """Initialize the ORCA data stream.

        Parameters
        ----------
        stream_name
            The ORCA filename. Only file streams are currently supported.
            Socket stream reading can be added later.
        rb_lib
            library of buffers for this stream.
        buffer_size
            length of tables to be read out in :meth:`read_chunk`.
        chunk_mode : 'any_full', 'only_full', or 'single_packet'
            sets the mode use for :meth:`read_chunk`.
        out_stream
            optional name of output stream for default `rb_lib` generation.

        Returns
        -------
        header_data
            a list of length 1 containing the raw buffer holding the ORCA header.
        """

        self.set_in_stream(stream_name)

        # read in the header
        packet = self.load_packet()
        if orca_packet.get_data_id(packet) != 0:
            raise RuntimeError(
                f"got data id {orca_packet.get_data_id(packet)} for header"
            )

        self.packet_id = 0
        self.any_full |= self.header_decoder.decode_packet(packet, self.packet_id)
        self.header = self.header_decoder.header

        # find the names of all decoders listed in the header AND in the rb_lib (if specified)
        decoder_names = self.header.get_decoder_list()
        if rb_lib is not None and "*" not in rb_lib:
            keep_decoders = []
            for name in decoder_names:
                if name in rb_lib:
                    keep_decoders.append(name)
            decoder_names = keep_decoders
            # check that all requested decoders are present
            for name in rb_lib.keys():
                if name not in keep_decoders:
                    log.warning(
                        f"decoder {name} (requested in rb_lib) not in data description in header"
                    )

        # get a mapping of data_ids-of-interest to instantiated decoders
        id_to_dec_name_dict = self.header.get_id_to_decoder_name_dict(
            shift_data_id=False
        )
        instantiated_decoders = {"OrcaHeaderDecoder": self.header_decoder}
        for data_id in id_to_dec_name_dict.keys():
            name = id_to_dec_name_dict[data_id]
            if name not in instantiated_decoders:
                if name not in globals():
                    log.warning(
                        f"no implementation of {name}, corresponding packets will be skipped"
                    )
                    continue
                decoder = globals()[name]
                instantiated_decoders[name] = decoder(header=self.header)
            self.decoder_id_dict[data_id] = instantiated_decoders[name]

        # initialize the buffers in rb_lib. Store them for fast lookup
        super().open_stream(
            stream_name,
            rb_lib,
            buffer_size=buffer_size,
            chunk_mode=chunk_mode,
            out_stream=out_stream,
        )
        if rb_lib is None:
            rb_lib = self.rb_lib
        good_buffers = []
        for data_id in self.decoder_id_dict.keys():
            name = id_to_dec_name_dict[data_id]
            if name not in self.rb_lib:
                log.info(f"skipping data from {name}")
                continue
            self.rbl_id_dict[data_id] = self.rb_lib[name]
            good_buffers.append(name)
        # check that we have instantiated decoders for all buffers
        for key in self.rb_lib:
            if key not in good_buffers:
                log.warning(f"buffer for {key} has no decoder")
        log.debug(f"rb_lib = {self.rb_lib}")

        # return header raw buffer
        if "OrcaHeaderDecoder" in rb_lib:
            header_rb_list = rb_lib["OrcaHeaderDecoder"]
            if len(header_rb_list) != 1:
                log.warning(
                    f"header_rb_list had length {len(header_rb_list)}, ignoring all but the first"
                )
            rb = header_rb_list[0]
        else:
            rb = RawBuffer(lgdo=self.header_decoder.make_lgdo())
        rb.lgdo.value = json.dumps(self.header)
        rb.loc = 1  # we have filled this buffer
        return [rb]

    def read_packet(self) -> bool:
        """Read a packet of data.

        Data written to the `rb_lib` attribute.
        """
        # read until we get a decodeable packet
        while True:
            packet = self.load_packet(skip_unknown_ids=True)
            if packet is None:
                return False
            self.packet_id += 1

            # look up the data id, decoder, and rbl
            data_id = orca_packet.get_data_id(packet, shift=False)
            log.debug(
                f"packet {self.packet_id}: data_id = {data_id}, decoder = {'None' if data_id not in self.decoder_id_dict else type(self.decoder_id_dict[data_id]).__name__}"
            )
            if data_id in self.rbl_id_dict:
                break

        # now decode
        decoder = self.decoder_id_dict[data_id]
        rbl = self.rbl_id_dict[data_id]
        self.any_full |= decoder.decode_packet(packet, self.packet_id, rbl)
        return True



import pygama.lgdo as lgdo


def test_datatype_name():
    struct = lgdo.Struct()
    assert struct.datatype_name() == "struct"


def test_form_datatype():
    struct = lgdo.Struct()
    assert struct.form_datatype() == "struct{}"


def test_update_datatype():
    struct = lgdo.Struct()
    struct.update_datatype()
    assert struct.attrs["datatype"] == "struct{}"


def test_init():
    obj_dict = {"scalar1": lgdo.Scalar(value=10)}
    attrs = {"attr1": 1}
    struct = lgdo.Struct(obj_dict=obj_dict, attrs=attrs)
    assert dict(struct) == obj_dict
    assert struct.attrs == attrs | {"datatype": "struct{scalar1}"}


def test_add_field():
    struct = lgdo.Struct()
    struct.add_field("scalar1", lgdo.Scalar(value=10))

    assert struct.attrs["datatype"] == "struct{scalar1}"
    assert struct["scalar1"].__class__.__name__ == "Scalar"

    struct.add_field("array1", lgdo.Array(shape=(700, 21), dtype="f", fill_val=2))
    assert struct.attrs["datatype"] == "struct{scalar1,array1}"


def test_remove_field():
    struct = lgdo.Struct()
    struct.add_field("scalar1", lgdo.Scalar(value=10))
    struct.add_field("array1", lgdo.Array(shape=(10), fill_val=0))

    struct.remove_field("scalar1")
    assert list(struct.keys()) == ["array1"]

    struct.remove_field("array1", delete=True)
    assert list(struct.keys()) == []



"""This module implements some helpers for setting up logging."""
import logging

import colorlog

DEBUG = logging.DEBUG
INFO = logging.INFO
WARNING = logging.WARNING
ERROR = logging.ERROR
FATAL = logging.FATAL
CRITICAL = logging.CRITICAL


def setup(level: int = logging.INFO, logger: logging.Logger = None) -> None:
    """Setup a colorful logging output.

    If `logger` is None, sets up only the ``pygama`` logger.

    Parameters
    ----------
    level
        logging level (see :mod:`logging` module).
    logger
        if not `None`, setup this logger.

    Examples
    --------
    >>> from pygama import logging
    >>> logging.setup(level=logging.DEBUG)
    """
    handler = colorlog.StreamHandler()
    handler.setFormatter(
        colorlog.ColoredFormatter("%(log_color)s%(name)s [%(levelname)s] %(message)s")
    )

    if logger is None:
        logger = colorlog.getLogger("pygama")

    logger.setLevel(level)
    logger.addHandler(handler)



import pytest

from pygama.lgdo.compression import RadwareSigcompress
from pygama.lgdo.compression.utils import str2wfcodec


def test_str2wfcodec():
    assert str2wfcodec("RadwareSigcompress()") == RadwareSigcompress()
    assert str2wfcodec("RadwareSigcompress(codec_shift=-32768)") == RadwareSigcompress(
        codec_shift=-32768
    )
    assert str2wfcodec(
        " RadwareSigcompress( codec_shift = -32768 ) "
    ) == RadwareSigcompress(codec_shift=-32768)

    with pytest.raises(ValueError):
        assert str2wfcodec("RadwareSigcompress") == RadwareSigcompress(
            codec_shift=-32768
        )
        assert str2wfcodec("RadwareSigcompress(blabla)") == RadwareSigcompress(
            codec_shift=-32768
        )



import json

import pygama.raw.raw_buffer as prb


def test_raw_buffer_list():
    rbl = prb.RawBufferList()
    rbl.set_from_json_dict(
        {
            "g{key:0>3d}": {
                "key_list": [[3, 6]],
                "out_stream": "$DATADIR/{file_key}_geds.lh5",
            }
        },
        kw_dict={"file_key": "run0"},
    )
    assert len(rbl) == 4
    assert rbl[1].out_stream == "$DATADIR/run0_geds.lh5"
    assert rbl.get_list_of("out_name") == ["g003", "g004", "g005", "g006"]
    assert rbl.get_list_of("key_list") == [[3], [4], [5], [6]]

    rbl.clear()
    rbl.set_from_json_dict(
        {
            "spms": {
                "key_list": [[3, 6]],
                "out_stream": "$DATADIR/{file_key}_geds.lh5",
                "out_name": "testspms",
            }
        },
        kw_dict={"file_key": "run0"},
    )
    assert rbl.get_list_of("out_name") == ["testspms"]
    assert rbl.get_list_of("key_list") == [[3, 4, 5, 6]]


def test_raw_buffer_lib_json_load():
    rb_json = """
    {
      "FCEventDecoder" : {
        "g{key:0>3d}" : {
          "key_list" : [[24, 64]],
          "out_stream" : "$DATADIR/{file_key}_geds.lh5:/geds"
        },
        "spms" : {
          "key_list" : [[6, 23]],
          "out_stream" : "$DATADIR/{file_key}_spms.lh5"
        },
        "puls" : {
          "key_list" : [0],
          "out_stream" : "$DATADIR/{file_key}_auxs.lh5:/auxs"
        },
        "muvt" : {
          "key_list" : [1, 5],
          "out_stream" : "$DATADIR/{file_key}_auxs.lh5:/auxs"
        }
      },
      "*" : {
        "{name}" : {
          "key_list" : ["*"],
          "out_stream" : "$DATADIR/{file_key}_others.lh5"
        }
      }
    }
    """
    json_dict = json.loads(rb_json)
    rblib = prb.RawBufferLibrary(json_dict=json_dict, kw_dict={"file_key": "run0"})
    rb_keyed = rblib["FCEventDecoder"].get_keyed_dict()
    name = rb_keyed[41].out_name
    assert name == "g041"



import pytest

import pygama.dsp.processors  # noqa: F401
from pygama.lgdo import LH5Store
from pygama.raw.build_raw import build_raw


@pytest.fixture(scope="session")
def geds_raw_tbl(lgnd_test_data):
    store = LH5Store()
    obj, _ = store.read_object(
        "/geds/raw",
        lgnd_test_data.get_path("lh5/LDQTA_r117_20200110T105115Z_cal_geds_raw.lh5"),
        n_rows=10,
    )
    return obj


@pytest.fixture(scope="session")
def spms_raw_tbl(lgnd_test_data):
    out_file = "/tmp/L200-comm-20211130-phy-spms.lh5"
    out_spec = {
        "FCEventDecoder": {
            "raw": {"key_list": [[0, 6]], "out_stream": f"{out_file}:/spms"}
        }
    }

    build_raw(
        in_stream=lgnd_test_data.get_path("fcio/L200-comm-20211130-phy-spms.fcio"),
        out_spec=out_spec,
        n_max=10,
        overwrite=True,
    )

    store = LH5Store()
    obj, _ = store.read_object("/spms/raw", out_file, n_rows=10)
    return obj



from __future__ import annotations

import numpy as np
from numba import guvectorize

from pygama.dsp.errors import DSPFatal
from pygama.dsp.utils import numba_defaults_kwargs as nb_kwargs


@guvectorize(
    [
        "void(float32[:], float32, float32, float32, float32[:])",
        "void(float64[:], float64, float64, float64, float64[:])",
    ],
    "(n),(),(),()->()",
    **nb_kwargs,
)
def time_point_thresh(
    w_in: np.ndarray, a_threshold: float, t_start: int, walk_forward: int, t_out: float
) -> None:
    """Find the index where the waveform value crosses the threshold, walking
    either forward or backward from the starting index.

    Parameters
    ----------
    w_in
        the input waveform.
    a_threshold
        the threshold value.
    t_start
        the starting index.
    walk_forward
        the backward (``0``) or forward (``1``) search direction.
    t_out
        the index where the waveform value crosses the threshold.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "tp_0": {
            "function": "time_point_thresh",
            "module": "pygama.dsp.processors",
            "args": ["wf_atrap", "bl_std", "tp_start", 0, "tp_0"],
            "unit": "ns"
        }
    """
    t_out[0] = np.nan

    if (
        np.isnan(w_in).any()
        or np.isnan(a_threshold)
        or np.isnan(t_start)
        or np.isnan(walk_forward)
    ):
        return

    if np.floor(t_start) != t_start:
        raise DSPFatal("The starting index must be an integer")

    if np.floor(walk_forward) != walk_forward:
        raise DSPFatal("The search direction must be an integer")

    if int(t_start) < 0 or int(t_start) >= len(w_in):
        raise DSPFatal("The starting index is out of range")

    if int(walk_forward) == 1:
        for i in range(int(t_start), len(w_in) - 1, 1):
            if w_in[i] <= a_threshold < w_in[i + 1]:
                t_out[0] = i
                return
    else:
        for i in range(int(t_start), 1, -1):
            if w_in[i - 1] < a_threshold <= w_in[i]:
                t_out[0] = i
                return


@guvectorize(
    [
        "void(float32[:], float32, float32, int64, char, float32[:])",
        "void(float64[:], float64, float64, int64, char, float64[:])",
    ],
    "(n),(),(),(),()->()",
    **nb_kwargs,
)
def interpolated_time_point_thresh(
    w_in: np.ndarray,
    a_threshold: float,
    t_start: int,
    walk_forward: int,
    mode_in: np.int8,
    t_out: float,
) -> None:
    """Find the time where the waveform value crosses the threshold

    Search performed walking either forward or backward from the starting
    index. Use interpolation to estimate a time between samples. Interpolation
    mode selected with `mode_in`.

    Parameters
    ----------
    w_in
        the input waveform.
    a_threshold
        the threshold value.
    t_start
        the starting index.
    walk_forward
        the backward (``0``) or forward (``1``) search direction.
    mode_in
        Character selecting which interpolation method to use. Note this
        must be passed as a ``int8``, e.g. ``ord('i')``. Options:

        * ``i`` -- integer `t_in`; equivalent to
          :func:`~.dsp.processors.fixed_sample_pickoff`
        * ``f`` -- floor; interpolated values are at previous neighbor, so
          threshold crossing is at next neighbor
        * ``c`` -- ceiling, interpolated values are at next neighbor, so
          threshold crossing is at previous neighbor
        * ``n`` -- nearest-neighbor interpolation; threshold crossing is
          half-way between samples
        * ``l`` -- linear interpolation
        * ``h`` -- Hermite cubic spline interpolation (*not implemented*)
        * ``s`` -- natural cubic spline interpolation (*not implemented*)
    t_out
        the index where the waveform value crosses the threshold.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "tp_0": {
            "function": "time_point_thresh",
            "module": "pygama.dsp.processors",
            "args": ["wf_atrap", "bl_std", "tp_start", 0, "'l'", "tp_0"],
            "unit": "ns"
        }
    """
    t_out[0] = np.nan

    if (
        np.isnan(w_in).any()
        or np.isnan(a_threshold)
        or np.isnan(t_start)
        or np.isnan(walk_forward)
    ):
        return

    if t_start < 0 or t_start >= len(w_in):
        return

    i_cross = -1
    if walk_forward > 0:
        for i in range(int(t_start), len(w_in) - 1, 1):
            if w_in[i] <= a_threshold < w_in[i + 1]:
                i_cross = i
    else:
        for i in range(int(t_start), 1, -1):
            if w_in[i - 1] < a_threshold <= w_in[i]:
                i_cross = i - 1

    if i_cross == -1:
        return

    if mode_in == ord("i"):  # return index before crossing
        t_out[0] = i_cross
    elif mode_in == ord("f"):  # return index before crossing
        t_out[0] = i_cross + 1
    elif mode_in == ord("c"):  # return index before crossing
        t_out[0] = i_cross
    elif mode_in == ord("n"):  # nearest-neighbor; return half-way between samps
        t_out[0] = i_cross + 0.5
    elif mode_in == ord("l"):  # linear
        t_out[0] = i_cross + (a_threshold - w_in[i_cross]) / (
            w_in[i_cross + 1] - w_in[i_cross]
        )
    else:
        raise DSPFatal("Unrecognized interpolation mode")



"""
- get_avse_cut (does AvsE)
- get_ae_cut (does A/E)
"""

import matplotlib.gridspec as gs
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import LogNorm

from pygama.math.histogram import get_bin_centers
from pygama.math.peak_fitting import *


def get_avse_cut(e_cal, current, plotFigure=None):
    # DEP range and nearby BG for BG subtraction
    dep_idxs = (e_cal > 1585) & (
        e_cal < 1595
    )  # (asymmetric around 1592 to account for CT)
    bg_idxs = (e_cal > 1560) & (e_cal < 1570)

    # SEP range and nearby BG for BG subtraction
    sep_idxs = (e_cal > 2090) & (e_cal < 2115)
    bg_sep_idxs = (e_cal > 2060) & (e_cal < 2080)

    # Bin the data in 2D, 25 keV bins
    # xedges = np.arange(200,2300,25)
    # e_cent = get_bin_centers(xedges)
    # y_max = np.zeros_like(e_cent)
    # plt.ion()
    # plt.figure()

    compton_shoulder = 2614 * (1 - 1.0 / (1 + 2 * 2614 / 511))

    # peaks = [238,510,583,727,860,1078,1512,1592,1806,compton_shoulder, 2614]
    # peaks = np.array(peaks)
    # e_cent = peaks
    # y_max = np.zeros_like(e_cent)

    xedges = np.arange(200, 2300, 25)
    e_cent = get_bin_centers(xedges)
    y_max = np.zeros_like(e_cent)

    # plt.ion()
    # plt.figure()
    for i, peak in enumerate(e_cent):
        plt.clf()
        # e_bin_idxs = (e_cal > peak-10) & (e_cal < peak+10 )
        e_bin_idxs = (e_cal > xedges[i]) & (e_cal < xedges[i + 1])
        a_ebin = current[e_bin_idxs]
        a_5 = np.percentile(a_ebin, 25)
        a_95 = np.percentile(a_ebin, 99)

        h, a_bins = np.histogram(a_ebin, bins=np.linspace(a_5, a_95, 500))

        a_bins_cent = get_bin_centers(a_bins)
        a_mode = a_bins_cent[np.argmax(h)]
        y_max[i] = a_mode

        p0 = get_gaussian_guess(h, a_bins_cent)
        fit_idxs = a_bins_cent > p0[0] - 5 * p0[1]
        p = fit_binned(gauss, h[fit_idxs], a_bins_cent[fit_idxs], p0)
        y_max[i] = p[0]

        # plt.plot(a_bins_cent,h,ls="steps")
        # plt.axvline(a_mode, c="r")
        # plt.title("Energy: {} keV".format(e_cent[i]))
        #
        # fit = gauss(a_bins_cent[fit_idxs], *p)
        # plt.plot(a_bins_cent[fit_idxs], fit, c="g")

        # guess = gauss(a_bins_cent[fit_idxs], *p0)
        # plt.plot(a_bins_cent[fit_idxs], guess, c="r")

        # inp = input("q to quit")
        # if inp == "q": exit()

    # quadratic fit
    # first fit a line:
    p_lin = np.polyfit(e_cent, y_max, 1)
    # find residuals:
    resid = y_max - np.poly1d(p_lin)(e_cent)
    resid_std = np.std(resid)
    # a really big residual is a pulser peak:
    puls_idx = resid > resid_std

    e_cent_cut = e_cent[~puls_idx]
    y_max_cut = y_max[~puls_idx]

    avse_quad = np.polyfit(e_cent_cut, y_max_cut, 2)
    avse_lin = np.polyfit(e_cent_cut, y_max_cut, 1)

    a_adjusted = current - np.poly1d(avse_quad)(e_cal)

    # Look at DEP, bg subtract the AvsE spectrum
    h_dep, bins = np.histogram(a_adjusted[dep_idxs], bins=5000)
    h_bg, bins = np.histogram(a_adjusted[bg_idxs], bins=bins)
    bin_centers = get_bin_centers(bins)
    h_bgs = h_dep - h_bg
    # fit AvsE peak to gaussian to get the 90% cut
    p0 = get_gaussian_guess(h_bgs, bin_centers)
    p = fit_binned(gauss, h_bgs, bin_centers, p0)
    fit = gauss(bin_centers, *p)

    ae_mean, ae_std = p[0], p[1]
    ae_cut = p[0] - 1.28 * p[1]  # cuts at 10% of CDF

    avse2, avse1, avse0 = avse_quad[:]
    avse_cut = ae_cut
    avse_mean = ae_mean
    avse_std = ae_std

    # plt.figure()
    # x = np.linspace(0,2700,5000)
    # plt.scatter(e_cent_cut, y_max_cut, color="k", s=10)
    # plt.scatter(e_cent[puls_idx], y_max[puls_idx], color="r", s=10)
    # plt.plot(x, np.poly1d(avse_quad)(x))
    # plt.plot(x, np.poly1d(avse_lin)(x))
    #
    # plt.figure()
    # xedges = np.arange(1000,2700,25)
    # aa_5 = np.percentile(a_adjusted,5)
    # aa_95 = np.percentile(a_adjusted,99)
    # yedges = np.linspace(aa_5, aa_95,1000)
    # H, xedges, yedges = np.histogram2d(e_cal, a_adjusted, bins=( xedges, yedges))
    # plt.imshow(H.T, interpolation='nearest', origin='low', aspect="auto",
    #             extent=[xedges[0], xedges[-1], yedges[0], yedges[-1]], cmap="OrRd", norm=LogNorm())
    # plt.scatter(e_cent_cut, y_max_cut - np.poly1d(avse_quad)(e_cent_cut) , color="k", s=10)
    # plt.axhline(ae_cut, color="k", ls="--")
    #
    # inp = input("q to quit")
    # if inp == "q": exit()

    if plotFigure is not None:
        ####
        # Plot A/E distributions
        ###
        plt.figure(plotFigure.number)
        plt.clf()
        grid = gs.GridSpec(2, 2)

        ax_dep = plt.subplot(grid[0, 0])
        ax_sep = plt.subplot(grid[1, 0])
        ax_ae = plt.subplot(grid[:, 1])

        # adjust a over e for mean
        a_over_e = a_adjusted
        ae_cut_mod = ae_cut

        h_dep, bins = np.histogram(
            a_over_e[dep_idxs], bins=np.linspace(-12 * ae_std, 8 * ae_std, 100)
        )
        h_bg, bins = np.histogram(a_over_e[bg_idxs], bins=bins)
        bin_centers = bins[:-1] + 0.5 * (bins[1] - bins[0])
        h_bgs = h_dep - h_bg

        h_sep, bins = np.histogram(a_over_e[sep_idxs], bins=bins)
        h_sepbg, bins = np.histogram(a_over_e[bg_sep_idxs], bins=bins)
        h_bgs_sep = h_sep - h_sepbg

        # ax_ae.plot(bin_centers,h_bgs / np.sum(h_bgs),  ls="steps-mid", color = "b", label = "DEP (BG subtracted)")
        # ax_ae.plot(bin_centers, h_bgs_sep/ np.sum(h_bgs), ls="steps-mid", color = "g", label = "SEP (BG subtracted)")

        ax_ae.plot(
            bin_centers, h_bgs, ls="steps-mid", color="b", label="DEP (BG subtracted)"
        )
        ax_ae.plot(
            bin_centers,
            h_bgs_sep,
            ls="steps-mid",
            color="g",
            label="SEP (BG subtracted)",
        )
        ax_ae.axvline(ae_cut_mod, color="r", ls=":")
        ax_ae.set_xlim(-12 * ae_std, 8 * ae_std)
        ax_ae.legend(loc=2)

        ax_ae.set_xlabel("A/E value [arb]")

        ###
        # Plot SEP/DEP before/after cut
        ##
        ae_cut_idxs = a_over_e > ae_cut_mod
        e_cal_aepass = e_cal[ae_cut_idxs]

        pad = 50
        bin_size = 0.2  # keV
        bins = np.arange(1592 - pad, 1592 + pad + bin_size, bin_size)

        ax_dep.hist(
            e_cal[(e_cal > 1592 - pad) & (e_cal < 1592 + pad)],
            histtype="step",
            color="k",
            label="DEP",
            bins=bins,
        )
        ax_dep.hist(
            e_cal_aepass[(e_cal_aepass > 1592 - pad) & (e_cal_aepass < 1592 + pad)],
            histtype="step",
            color="b",
            label="After Cut",
            bins=bins,
        )
        ax_dep.legend(loc=2)
        ax_dep.set_xlabel("Energy [keV]")

        bins = np.arange(2103 - pad, 2103 + pad + bin_size, bin_size)
        ax_sep.hist(
            e_cal[(e_cal > 2103 - pad) & (e_cal < 2103 + pad)],
            histtype="step",
            color="k",
            label="SEP",
            bins=bins,
        )
        ax_sep.hist(
            e_cal_aepass[(e_cal_aepass > 2103 - pad) & (e_cal_aepass < 2103 + pad)],
            histtype="step",
            color="b",
            label="After Cut",
            bins=bins,
        )
        ax_sep.legend(loc=2)
        ax_sep.set_xlabel("Energy [keV]")

    return avse2, avse1, avse0, avse_cut, avse_mean, avse_std


def get_ae_cut(e_cal, current, plotFigure=None):
    # try to get a rough A/E cut

    # DEP range and nearby BG for BG subtraction
    dep_idxs = (e_cal > 1585) & (
        e_cal < 1595
    )  # (asymmetric around 1592 to account for CT)
    bg_idxs = (e_cal > 1560) & (e_cal < 1570)

    # SEP range and nearby BG for BG subtraction
    sep_idxs = (e_cal > 2090) & (e_cal < 2115)
    bg_sep_idxs = (e_cal > 2060) & (e_cal < 2080)

    a_over_e = current / e_cal

    # # peaks = [2381]
    # peaks = [1512, 1592,1620,1806,2381]
    # ae_cents = np.zeros((len(peaks)))

    h_dep, bins = np.histogram(a_over_e[dep_idxs], bins=500)
    h_bg, bins = np.histogram(a_over_e[bg_idxs], bins=bins)
    bin_centers = get_bin_centers(bins)
    h_bgs = h_dep - h_bg

    p0 = get_gaussian_guess(h_bgs, bin_centers)
    p = fit_binned(gauss, h_bgs, bin_centers, p0)
    fit = gauss(bin_centers, *p)

    ae_mean, ae_std = p[0], p[1]
    ae_cut = p[0] - 1.28 * p[1]  # cuts at 10% of CDF

    if plotFigure is not None:
        ####
        # Plot A/E distributions
        ###
        plt.figure(plotFigure.number)
        plt.clf()
        grid = gs.GridSpec(2, 2)

        ax_dep = plt.subplot(grid[0, 0])
        ax_sep = plt.subplot(grid[1, 0])
        ax_ae = plt.subplot(grid[:, 1])

        # adjust a over e for mean

        a_over_e = (a_over_e - ae_mean) / ae_std
        ae_cut_mod = (ae_cut - ae_mean) / ae_std

        h_dep, bins = np.histogram(a_over_e[dep_idxs], bins=np.linspace(-8, 6, 50))
        h_bg, bins = np.histogram(a_over_e[bg_idxs], bins=bins)
        bin_centers = bins[:-1] + 0.5 * (bins[1] - bins[0])
        h_bgs = h_dep - h_bg

        h_sep, bins = np.histogram(a_over_e[sep_idxs], bins=bins)
        h_sepbg, bins = np.histogram(a_over_e[bg_sep_idxs], bins=bins)
        h_bgs_sep = h_sep - h_sepbg

        ax_ae.plot(
            bin_centers,
            h_bgs / np.sum(h_bgs),
            ls="steps-mid",
            color="b",
            label="DEP (BG subtracted)",
        )
        ax_ae.plot(
            bin_centers,
            h_bgs_sep / np.sum(h_bgs),
            ls="steps-mid",
            color="g",
            label="SEP (BG subtracted)",
        )
        # plt.plot(bin_centers, fit, color="g")
        ax_ae.axvline(ae_cut_mod, color="r", ls=":")
        ax_ae.set_xlim(-8, 5)
        ax_ae.legend(loc=2)

        ax_ae.set_xlabel("A/E value [arb]")

        ###
        # Plot SEP/DEP before/after cut
        ##
        ae_cut_idxs = a_over_e > ae_cut_mod
        e_cal_aepass = e_cal[ae_cut_idxs]

        pad = 50
        bins = np.linspace(1592 - pad, 1592 + pad, 2 * pad + 1)

        ax_dep.hist(
            e_cal[(e_cal > 1592 - pad) & (e_cal < 1592 + pad)],
            histtype="step",
            color="k",
            label="DEP",
            bins=bins,
        )
        ax_dep.hist(
            e_cal_aepass[(e_cal_aepass > 1592 - pad) & (e_cal_aepass < 1592 + pad)],
            histtype="step",
            color="b",
            label="After Cut",
            bins=bins,
        )
        ax_dep.legend(loc=2)
        ax_dep.set_xlabel("Energy [keV]")

        bins = np.linspace(2103 - pad, 2103 + pad, 2 * pad + 1)
        ax_sep.hist(
            e_cal[(e_cal > 2103 - pad) & (e_cal < 2103 + pad)],
            histtype="step",
            color="k",
            label="SEP",
            bins=bins,
        )
        ax_sep.hist(
            e_cal_aepass[(e_cal_aepass > 2103 - pad) & (e_cal_aepass < 2103 + pad)],
            histtype="step",
            color="b",
            label="After Cut",
            bins=bins,
        )
        ax_sep.legend(loc=2)
        ax_sep.set_xlabel("Energy [keV]")

    return ae_cut, ae_mean, ae_std



import pytest

from pygama import lgdo
from pygama.dsp.processing_chain import build_processing_chain


def test_waveform_slicing(geds_raw_tbl):
    dsp_config = {
        "outputs": ["wf_blsub"],
        "processors": {
            "wf_blsub": {
                "function": "bl_subtract",
                "module": "pygama.dsp.processors",
                "args": ["waveform[0:100]", "baseline", "wf_blsub"],
                "unit": "ADC",
            },
        },
    }
    proc_chain, _, tbl_out = build_processing_chain(geds_raw_tbl, dsp_config)
    proc_chain.execute(0, 1)

    assert list(tbl_out.keys()) == ["wf_blsub"]
    assert isinstance(tbl_out["wf_blsub"], lgdo.WaveformTable)
    assert tbl_out["wf_blsub"].wf_len == 100


def test_processor_none_arg(geds_raw_tbl):
    dsp_config = {
        "outputs": ["wf_cum"],
        "processors": {
            "wf_cum": {
                "function": "cumsum",
                "module": "numpy",
                "args": ["waveform", 1, None, "wf_cum"],
                "kwargs": {"signature": "(n),(),()->(n)", "types": ["fii->f"]},
                "unit": "ADC",
            }
        },
    }
    proc_chain, _, _ = build_processing_chain(geds_raw_tbl, dsp_config)
    proc_chain.execute(0, 1)

    dsp_config["processors"]["wf_cum"]["args"][2] = "None"
    proc_chain, _, _ = build_processing_chain(geds_raw_tbl, dsp_config)
    proc_chain.execute(0, 1)


def test_processor_kwarg_assignment(geds_raw_tbl):
    dsp_config = {
        "outputs": ["wf_cum"],
        "processors": {
            "wf_cum": {
                "function": "cumsum",
                "module": "numpy",
                "args": ["waveform", "axis=1", "out=wf_cum"],
                "kwargs": {"signature": "(n),()->(n)", "types": ["fi->f"]},
                "unit": "ADC",
            }
        },
    }
    proc_chain, _, _ = build_processing_chain(geds_raw_tbl, dsp_config)
    proc_chain.execute(0, 1)

    dsp_config["processors"]["wf_cum"]["args"][1] = "dtypo=None"
    proc_chain, _, _ = build_processing_chain(geds_raw_tbl, dsp_config)
    with pytest.raises(TypeError):
        proc_chain.execute(0, 1)


def test_processor_dtype_arg(geds_raw_tbl):
    dsp_config = {
        "outputs": ["wf_cum"],
        "processors": {
            "wf_cum": {
                "function": "cumsum",
                "module": "numpy",
                "args": ["waveform", "axis=0", "dtype='int32'", "out=wf_cum"],
                "kwargs": {"signature": "(n),(),()->(n)", "types": ["fiU->i"]},
                "unit": "ADC",
            }
        },
    }
    proc_chain, _, _ = build_processing_chain(geds_raw_tbl, dsp_config)
    proc_chain.execute(0, 1)


def test_scipy_gauss_filter(geds_raw_tbl):
    dsp_config = {
        "outputs": ["wf_gaus"],
        "processors": {
            "wf_gaus": {
                "function": "gaussian_filter1d",
                "module": "scipy.ndimage",
                "args": [
                    "waveform",
                    "0.1*us",
                    "mode='reflect'",
                    "truncate=3",
                    "output=wf_gaus",
                ],
                "kwargs": {"signature": "(n),(),(),()->(n)", "types": ["ffUf->f"]},
                "unit": "ADC",
            }
        },
    }
    proc_chain, _, _ = build_processing_chain(geds_raw_tbl, dsp_config)
    proc_chain.execute(0, 1)


def test_histogram_processor_fixed_witdth(spms_raw_tbl):
    dsp_config = {
        "outputs": ["hist_weights", "hist_borders"],
        "processors": {
            "hist_weights , hist_borders": {
                "function": "histogram",
                "module": "pygama.dsp.processors.histogram",
                "args": ["waveform", "hist_weights(100)", "hist_borders(101)"],
                "unit": ["none", "ADC"],
            }
        },
    }
    proc_chain, _, _ = build_processing_chain(spms_raw_tbl, dsp_config)
    proc_chain.execute(0, 1)


def test_processor_variable_array_output(spms_raw_tbl):
    dsp_config = {
        "outputs": ["vt_max_out"],
        "processors": {
            "vt_max_out, vt_min_out, n_max_out, n_min_out": {
                "function": "get_multi_local_extrema",
                "module": "pygama.dsp.processors",
                "args": [
                    "waveform",
                    5,
                    5,
                    0,
                    10,
                    0,
                    "vt_max_out(10)",
                    "vt_min_out(10)",
                    "n_max_out",
                    "n_min_out",
                ],
                "unit": "ADC",
            }
        },
    }

    proc_chain, _, _ = build_processing_chain(spms_raw_tbl, dsp_config)
    proc_chain.execute(0, 1)



from __future__ import annotations

import numpy as np
from numba import guvectorize

from pygama.dsp.utils import numba_defaults_kwargs as nb_kwargs


@guvectorize(
    [
        "void(float32[:], float32[:], float32[:], float32[:], float32[:])",
        "void(float64[:], float64[:], float64[:], float64[:], float64[:])",
    ],
    "(n)->(),(),(),()",
    **nb_kwargs,
)
def min_max(
    w_in: np.ndarray, t_min: int, t_max: int, a_min: float, a_max: float
) -> None:
    """Find the value and index of the minimum and maximum values
    in the waveform.

    Note
    ----
    The first found instance of each extremum in the waveform will be returned.

    Parameters
    ----------
    w_in
        the input waveform
    t_min
        the index of the minimum value
    t_max
        the index of the maximum value
    a_min
        the minimum value
    a_max
        the maximum value

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "tp_min, tp_max, wf_min, wf_max": {
            "function": "min_max",
            "module": "pygama.dsp.processors",
            "args": ["waveform", "tp_min", "tp_max", "wf_min", "wf_max"],
            "unit": ["ns", "ns", "ADC", "ADC"]
        }
    """
    a_min[0] = np.nan
    a_max[0] = np.nan
    t_min[0] = np.nan
    t_max[0] = np.nan

    if np.isnan(w_in).any():
        return

    min_index = 0
    max_index = 0

    for i in range(0, len(w_in), 1):
        if w_in[i] < w_in[min_index]:
            min_index = i
        if w_in[i] > w_in[max_index]:
            max_index = i

    a_min[0] = w_in[min_index]
    a_max[0] = w_in[max_index]
    t_min[0] = float(min_index)
    t_max[0] = float(max_index)



from __future__ import annotations

import numpy as np
from numba import guvectorize

from pygama.dsp.errors import DSPFatal
from pygama.dsp.utils import numba_defaults_kwargs as nb_kwargs


@guvectorize(
    ["void(float32[:], float32, float32[:])", "void(float64[:], float64, float64[:])"],
    "(n),()->(n)",
    **nb_kwargs,
)
def moving_window_left(w_in: np.ndarray, length: float, w_out: np.ndarray) -> None:
    """Applies a moving average window to the waveform.

    Note
    ----
    Starts from the left and assumes that the baseline is at zero.

    Parameters
    ----------
    w_in
        the input waveform.
    length
        length of the moving window to be applied.
    w_out
        output waveform after moving window applied.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "wf_mw": {
            "function": "moving_window_left",
            "module": "pygama.dsp.processors",
            "args": ["wf_pz", "96*ns", "wf_mw"],
            "unit": "ADC"
        }
    """

    w_out[:] = np.nan

    if np.isnan(w_in).any():
        return

    if not length >= 0 or not length < len(w_in):
        raise DSPFatal(
            "length is out of range, must be between 0 and the length of the waveform"
        )

    w_out[0] = w_in[0]
    for i in range(1, int(length)):
        w_out[i] = w_out[i - 1] + (w_in[i] - w_in[0]) / length
    for i in range(int(length), len(w_in)):
        w_out[i] = w_out[i - 1] + (w_in[i] - w_in[i - int(length)]) / length


@guvectorize(
    ["void(float32[:], float32, float32[:])", "void(float64[:], float64, float64[:])"],
    "(n),()->(n)",
    **nb_kwargs,
)
def moving_window_right(w_in: np.ndarray, length: float, w_out: np.ndarray) -> None:
    """Applies a moving average window to the waveform from the right.

    Parameters
    ----------
    w_in
        the input waveform.
    length
        length of the moving window to be applied.
    w_out
        output waveform after moving window applied.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "wf_mw": {
            "function": "moving_window_right",
            "module": "pygama.dsp.processors",
            "args": ["wf_pz", "96*ns", "wf_mw"],
            "unit": "ADC"
        }
    """

    w_out[:] = np.nan

    if np.isnan(w_in).any():
        return

    if not length >= 0 or not length < len(w_in):
        raise DSPFatal(
            "length is out of range, must be between 0 and the length of the waveform"
        )

    w_out[-1] = w_in[-1]
    for i in range(1, int(length), 1):
        w_out[len(w_in) - 1 - i] = (
            w_out[len(w_in) - i] + (w_in[len(w_in) - 1 - i] - w_out[-1]) / length
        )
    for i in range(int(length), len(w_in), 1):
        w_out[len(w_in) - 1 - i] = (
            w_out[len(w_in) - i]
            + (w_in[len(w_in) - 1 - i] - w_in[len(w_in) - 1 - i + int(length)]) / length
        )


@guvectorize(
    [
        "void(float32[:], float32, float32, int32, float32[:])",
        "void(float64[:], float64, float64, int32, float64[:])",
    ],
    "(n),(),(),()->(n)",
    **nb_kwargs,
)
def moving_window_multi(
    w_in: np.ndarray, length: float, num_mw: int, mw_type: int, w_out: np.ndarray
) -> None:
    """Apply a series of moving-average windows to the waveform, alternating
    its application between the left and the right.

    Parameters
    ----------
    w_in
        the input waveform.
    length
        length of the moving window to be applied.
    num_mw
        the number of moving windows.
    mw_type
        - ``0`` -- alternate moving windows right and left
        - ``1`` -- only left
        - ``2`` -- only right
    w_out
        the windowed waveform.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "curr_av": {
            "function": "moving_window_multi",
            "module": "pygama.dsp.processors",
            "args": ["curr", "96*ns", "3", "0", "curr_av"],
            "unit": "ADC/sample"
        }
    """
    w_out[:] = np.nan

    if np.isnan(w_in).any():
        return

    if np.floor(length) != length:
        raise DSPFatal("The length of the moving window must be an integer")

    if np.floor(num_mw) != num_mw:
        raise DSPFatal("The number of moving windows must be an integer")

    if int(length) < 0 or int(length) >= len(w_in):
        raise DSPFatal("The length of the moving window is out of range")

    if int(num_mw) < 0:
        raise DSPFatal("The number of moving windows much be positive")

    w_buf = w_in.copy()
    for i in range(0, int(num_mw), 1):
        if ((i % 2 == 1) & (mw_type == 0)) | (mw_type == 2):
            w_out[-1] = w_buf[-1]
            for i in range(1, int(length), 1):
                w_out[len(w_buf) - 1 - i] = (
                    w_out[len(w_buf) - i]
                    + (w_buf[len(w_buf) - 1 - i] - w_out[-1]) / length
                )
            for i in range(int(length), len(w_buf), 1):
                w_out[len(w_buf) - 1 - i] = (
                    w_out[len(w_buf) - i]
                    + (
                        w_buf[len(w_buf) - 1 - i]
                        - w_buf[len(w_buf) - 1 - i + int(length)]
                    )
                    / length
                )
        else:
            w_out[0] = w_buf[0]
            for i in range(1, int(length)):
                w_out[i] = w_out[i - 1] + (w_buf[i] - w_buf[0]) / length
            for i in range(int(length), len(w_buf)):
                w_out[i] = w_out[i - 1] + (w_buf[i] - w_buf[i - int(length)]) / length
        w_buf = w_out.copy()


@guvectorize(
    ["void(float32[:], float32, float32[:])", "void(float64[:], float64, float64[:])"],
    "(n),(),(m)",
    **nb_kwargs,
)
def avg_current(w_in: np.ndarray, length: float, w_out: np.ndarray) -> None:
    """Calculate the derivative of a waveform, averaged across `length` samples.

    Parameters
    ----------
    w_in
        the input waveform.
    length
        length of the moving window to be applied.
    w_out
        output waveform after derivation.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "curr": {
            "function": "avg_current",
            "module": "pygama.dsp.processors",
            "args": ["wf_pz", 1, "curr(len(wf_pz)-1, f)"],
            "unit": "ADC/sample"
        }
    """

    w_out[:] = np.nan

    if np.isnan(w_in).any():
        return

    if not length >= 0 or not length < len(w_in):
        raise DSPFatal(
            "length is out of range, must be between 0 and the length of the waveform"
        )

    w_out[:] = w_in[int(length) :] - w_in[: -int(length)]
    w_out /= length



import os
from pathlib import Path

import pytest

from pygama.lgdo.lh5_store import LH5Store, ls
from pygama.raw import build_raw

config_dir = Path(__file__).parent / "configs"


def test_build_raw_basics(lgnd_test_data):
    with pytest.raises(FileNotFoundError):
        build_raw(in_stream="non-existent-file")

    with pytest.raises(FileNotFoundError):
        build_raw(
            in_stream=lgnd_test_data.get_path("fcio/L200-comm-20211130-phy-spms.fcio"),
            out_spec="non-existent-file.json",
        )


def test_build_raw_fc(lgnd_test_data):
    build_raw(
        in_stream=lgnd_test_data.get_path("fcio/L200-comm-20211130-phy-spms.fcio"),
        overwrite=True,
    )

    assert lgnd_test_data.get_path("fcio/L200-comm-20211130-phy-spms.lh5") != ""

    out_file = "/tmp/L200-comm-20211130-phy-spms.lh5"

    build_raw(
        in_stream=lgnd_test_data.get_path("fcio/L200-comm-20211130-phy-spms.fcio"),
        out_spec=out_file,
        overwrite=True,
    )

    assert os.path.exists("/tmp/L200-comm-20211130-phy-spms.lh5")


def test_build_raw_fc_out_spec(lgnd_test_data):
    out_file = "/tmp/L200-comm-20211130-phy-spms.lh5"
    out_spec = {
        "FCEventDecoder": {"spms": {"key_list": [[2, 4]], "out_stream": out_file}}
    }

    build_raw(
        in_stream=lgnd_test_data.get_path("fcio/L200-comm-20211130-phy-spms.fcio"),
        out_spec=out_spec,
        n_max=10,
        overwrite=True,
    )

    store = LH5Store()
    lh5_obj, n_rows = store.read_object("/spms", out_file)
    assert n_rows == 10
    assert (lh5_obj["channel"].nda == [2, 3, 4, 2, 3, 4, 2, 3, 4, 2]).all()

    build_raw(
        in_stream=lgnd_test_data.get_path("fcio/L200-comm-20211130-phy-spms.fcio"),
        out_spec=f"{config_dir}/fc-out-spec.json",
        n_max=10,
        overwrite=True,
    )


def test_build_raw_fc_channelwise_out_spec(lgnd_test_data):
    out_file = "/tmp/L200-comm-20211130-phy-spms.lh5"
    out_spec = {
        "FCEventDecoder": {
            "ch{key}": {
                "key_list": [[0, 6]],
                "out_stream": out_file + ":{name}",
                "out_name": "raw",
            }
        }
    }

    build_raw(
        in_stream=lgnd_test_data.get_path("fcio/L200-comm-20211130-phy-spms.fcio"),
        out_spec=out_spec,
        overwrite=True,
    )

    assert ls(out_file) == ["ch0", "ch1", "ch2", "ch3", "ch4", "ch5"]
    assert ls(out_file, "ch0/") == ["ch0/raw"]
    assert ls(out_file, "ch0/raw/waveform") == ["ch0/raw/waveform"]


def test_build_raw_orca(lgnd_test_data):
    build_raw(
        in_stream=lgnd_test_data.get_path("orca/fc/L200-comm-20220519-phy-geds.orca"),
        overwrite=True,
    )

    assert lgnd_test_data.get_path("orca/fc/L200-comm-20220519-phy-geds.lh5") != ""

    out_file = "/tmp/L200-comm-20220519-phy-geds.lh5"

    build_raw(
        in_stream=lgnd_test_data.get_path("orca/fc/L200-comm-20220519-phy-geds.orca"),
        out_spec=out_file,
        overwrite=True,
    )

    assert os.path.exists("/tmp/L200-comm-20220519-phy-geds.lh5")


def test_build_raw_orca_out_spec(lgnd_test_data):
    out_file = "/tmp/L200-comm-20220519-phy-geds.lh5"
    out_spec = {
        "ORFlashCamADCWaveformDecoder": {
            "geds": {"key_list": [[1028802, 1028804]], "out_stream": out_file}
        }
    }

    build_raw(
        in_stream=lgnd_test_data.get_path("orca/fc/L200-comm-20220519-phy-geds.orca"),
        out_spec=out_spec,
        n_max=10,
        overwrite=True,
    )

    store = LH5Store()
    lh5_obj, n_rows = store.read_object("/geds", out_file)
    assert n_rows == 10
    assert (lh5_obj["channel"].nda == [2, 3, 4, 2, 3, 4, 2, 3, 4, 2]).all()

    build_raw(
        in_stream=lgnd_test_data.get_path("orca/fc/L200-comm-20220519-phy-geds.orca"),
        out_spec=f"{config_dir}/orca-out-spec.json",
        n_max=10,
        overwrite=True,
    )


def test_build_raw_compass(lgnd_test_data):
    build_raw(
        in_stream=lgnd_test_data.get_path("compass/compass_test_data.BIN"),
        overwrite=True,
        compass_config_file=lgnd_test_data.get_path(
            "compass/compass_test_data_settings.xml"
        ),
    )

    assert lgnd_test_data.get_path("compass/compass_test_data.lh5") != ""

    out_file = "/tmp/compass_test_data.lh5"

    build_raw(
        in_stream=lgnd_test_data.get_path("compass/compass_test_data.BIN"),
        out_spec=out_file,
        overwrite=True,
        compass_config_file=lgnd_test_data.get_path(
            "compass/compass_test_data_settings.xml"
        ),
    )

    assert os.path.exists("/tmp/compass_test_data.lh5")


def test_build_raw_compass_out_spec(lgnd_test_data):
    out_file = "/tmp/compass_test_data.lh5"
    out_spec = {
        "CompassEventDecoder": {"spms": {"key_list": [[0, 1]], "out_stream": out_file}}
    }

    build_raw(
        in_stream=lgnd_test_data.get_path("compass/compass_test_data.BIN"),
        out_spec=out_spec,
        n_max=10,
        overwrite=True,
        compass_config_file=lgnd_test_data.get_path(
            "compass/compass_test_data_settings.xml"
        ),
    )

    store = LH5Store()
    lh5_obj, n_rows = store.read_object("/spms", out_file)
    assert n_rows == 10
    assert (lh5_obj["channel"].nda == [0, 1, 0, 1, 0, 1, 0, 1, 0, 1]).all()


def test_build_raw_compass_out_spec_no_config(lgnd_test_data):
    out_file = "/tmp/compass_test_data.lh5"
    out_spec = {
        "CompassEventDecoder": {"spms": {"key_list": [[0, 1]], "out_stream": out_file}}
    }

    build_raw(
        in_stream=lgnd_test_data.get_path("compass/compass_test_data.BIN"),
        out_spec=out_spec,
        n_max=10,
        overwrite=True,
    )

    store = LH5Store()
    lh5_obj, n_rows = store.read_object("/spms", out_file)
    assert n_rows == 10
    assert (lh5_obj["channel"].nda == [0, 1, 0, 1, 0, 1, 0, 1, 0, 1]).all()


def test_build_raw_overwrite(lgnd_test_data):
    with pytest.raises(FileExistsError):
        build_raw(
            in_stream=lgnd_test_data.get_path("fcio/L200-comm-20211130-phy-spms.fcio")
        )


def test_build_raw_orca_sis3316(lgnd_test_data):
    out_file = "/tmp/coherent-run1141-bkg.lh5"
    out_spec = {
        "ORSIS3316WaveformDecoder": {
            "Card1": {"key_list": [48], "out_stream": out_file}
        }
    }

    build_raw(
        in_stream=lgnd_test_data.get_path("orca/sis3316/coherent-run1141-bkg.orca"),
        out_spec=out_spec,
        n_max=10,
        overwrite=True,
    )

    assert os.path.exists(out_file)
